# -*- coding: utf-8 -*-
"""
Tests for abagen.datasets.fetchers module
"""

from pathlib import Path

import pytest

from abagen.datasets import fetchers


def test_fetch_microarray():
    # test different `donors` inputs
    f1 = fetchers.fetch_microarray(donors=['12876'])
    f2 = fetchers.fetch_microarray(donors='12876')
    f3 = fetchers.fetch_microarray(donors='H0351.1009')
    f4 = fetchers.fetch_microarray(donors=None)

    # test n_proc
    fetchers.fetch_microarray(donors=['12876', '15496'], n_proc=2)

    # don't test this -- it will take a wicked long time
    # f5 = datasets.fetch_microarray(donors='all')

    assert f1 == f2 == f3 == f4
    assert len(f1) == 1  # only one donor
    for k in ['microarray', 'annotation', 'pacall', 'probes', 'ontology']:
        assert k in f1['12876']

    # check downloading incorrect donor
    with pytest.raises(ValueError):
        fetchers.fetch_microarray(donors='notadonor')

    with pytest.raises(ValueError):
        fetchers.fetch_microarray(donors=['notadonor'])


def test_fetch_raw_mri():
    # test different `donors` inputs
    f1 = fetchers.fetch_raw_mri(donors=['12876'])
    f2 = fetchers.fetch_raw_mri(donors='12876')
    f3 = fetchers.fetch_raw_mri(donors='H0351.1009')
    f4 = fetchers.fetch_raw_mri(donors=None)

    assert f1 == f2 == f3 == f4
    assert len(f1) == 1
    for k in ['t1w', 't2w']:
        assert k in f1['12876']

    # check downloading incorrect donor
    with pytest.raises(ValueError):
        fetchers.fetch_raw_mri(donors='notadonor')

    with pytest.raises(ValueError):
        fetchers.fetch_raw_mri(donors=['notadonor'])


def test_fetch_rnaseq():
    f1 = fetchers.fetch_rnaseq(donors=['9861'])
    f2 = fetchers.fetch_rnaseq(donors='9861')
    f3 = fetchers.fetch_rnaseq(donors='H0351.2001')
    f4 = fetchers.fetch_rnaseq(donors=None)

    assert f1 == f2 == f3 == f4
    assert len(f1) == 1
    for k in ['genes', 'ontology', 'counts', 'tpm', 'annotation']:
        assert k in f1['9861']

    with pytest.raises(ValueError):
        fetchers.fetch_rnaseq(donors='notadonor')

    with pytest.raises(ValueError):
        fetchers.fetch_rnaseq(donors=['9861', 'notadonor'])


@pytest.mark.parametrize('group, expected', [
    ('brain', 2413),
    ('neuron', 2530),
    ('oligodendrocyte', 1769),
    ('synaptome', 1886),
    ('layers', 46)
])
def test_get_gene_group(group, expected):
    assert len(fetchers.fetch_gene_group(group)) == expected

    with pytest.raises(ValueError):
        fetchers.fetch_gene_group('notagroup')


def test_fetch_desikan_killiany():
    atlas = fetchers.fetch_desikan_killiany()
    assert sorted(atlas) == ['image', 'info']
    img, info = Path(atlas['image']), Path(atlas['info'])
    assert img.name == 'atlas-desikankilliany.nii.gz' and img.exists()
    assert info.name == 'atlas-desikankilliany.csv' and info.exists()

    atlas = fetchers.fetch_desikan_killiany(native=True)
    donors = ['10021', '12876', '14380', '15496', '15697', '9861']
    assert sorted(atlas) == ['image', 'info']
    assert sorted(atlas['image']) == donors
    for donor in donors:
        img = Path(atlas['image'][donor])
        assert img.name == 'atlas-desikankilliany.nii.gz' and img.exists()
    info = Path(atlas['info'])
    assert info.name == 'atlas-desikankilliany.csv' and info.exists()



"""
Functions for fetching data relevant to the Allen Brain Atlas human microarray
dataset
"""

__all__ = [
    'fetch_microarray', 'fetch_raw_mri', 'fetch_desikan_killiany',
    'fetch_gene_group', 'fetch_rnaseq', 'fetch_donor_info', 'fetch_freesurfer',
    'WELL_KNOWN_IDS', '_get_dataset_dir'
]

from .fetchers import (fetch_microarray, fetch_raw_mri, fetch_desikan_killiany,
                       fetch_gene_group, fetch_rnaseq, fetch_donor_info,
                       fetch_freesurfer, WELL_KNOWN_IDS)
from .utils import _get_dataset_dir



# -*- coding: utf-8 -*-
"""
Tests for abagen.io module
"""

import os.path as op

import numpy as np
import pandas as pd
import pytest

from abagen import io
from abagen.utils import flatten_dict


annot_cols = [
    'structure_id', 'slab_num', 'well_id', 'slab_type', 'structure_acronym',
    'structure_name', 'polygon_id', 'mri_voxel_x', 'mri_voxel_y',
    'mri_voxel_z', 'mni_x', 'mni_y', 'mni_z'
]

ont_cols = [
    'id', 'acronym', 'name', 'parent_structure_id', 'hemisphere',
    'graph_order', 'structure_id_path', 'color_hex_triplet'
]

probe_cols = [
    'probe_name', 'gene_id', 'gene_symbol', 'gene_name', 'entrez_id',
    'chromosome'
]
gene_cols = [
    'gene_id', 'entrez_id', 'chromosome', 'strand', 'number_of_transcripts',
    'median_transcriptome_length', 'median_genome_length',
    'median_number_of_exons', 'median_gene_start', 'median_gene_end'
]
rna_annot_cols = [
    'RNAseq_sample_name', 'replicate_sample', 'sample_name', 'well_id',
    'microarray_run_id', 'ontology_color', 'main_structure',
    'sub_structure', 'structure_id', 'structure_acronym', 'hemisphere',
    'brain', 'million_clusters', 'clip_percentage', 'RIN_RNA_quality',
    'rnaseq_run_id', 'A.Pct', 'C.Pct', 'G.Pct', 'T.Pct', 'N.Pct'
]


@pytest.mark.parametrize('key, has_parq, columns', [
    ('microarray', True, None),
    ('pacall', True, None),
    ('annotation', False, annot_cols),
    ('ontology', False, ont_cols),
    ('probes', False, probe_cols),
])
def test_readfiles(testfiles, key, has_parq, columns):
    for d, fn in flatten_dict(testfiles, key).items():
        func = getattr(io, 'read_{}'.format(key))

        # check file (CSV + parquet) exist
        assert op.exists(fn)
        if has_parq and io.use_parq:
            assert op.exists(fn.rpartition('.csv')[0] + '.parq')

        # check loading from filepath
        data = func(fn, parquet=True) if has_parq else func(fn)
        assert isinstance(data, pd.DataFrame)

        # check loading from dataframe (should return same object)
        data2 = func(data)
        assert id(data) == id(data2)

        # check that copy parameter works as expected
        data3 = func(data, copy=True)
        assert isinstance(data3, pd.DataFrame)
        assert id(data) != id(data3)

        # confirm columns are as expected
        if columns is not None:
            assert np.all(columns == data.columns)

        # confirm errors
        with pytest.raises(TypeError):
            func(1)

        with pytest.raises(TypeError):
            func([1, 2, 3])

        with pytest.raises(FileNotFoundError):
            func('notafile')


@pytest.mark.parametrize('key, columns', [
    ('genes', gene_cols),
    ('ontology', ont_cols),
    ('counts', None),
    ('tpm', None),
    ('annotation', rna_annot_cols)
])
def test_readrnaseq(rnafiles, key, columns):
    for d, fn in flatten_dict(rnafiles, key).items():
        func = getattr(io, 'read_{}'.format(key))

        # check file exists
        assert op.exists(fn)

        # check loading from filepath
        data = func(fn)
        assert isinstance(data, pd.DataFrame)

        # check loading from dataframe (should return same object)
        data2 = func(data)
        assert id(data) == id(data2)

        # check that copy parameter works as expected
        data3 = func(data, copy=True)
        assert isinstance(data3, pd.DataFrame)
        assert id(data) != id(data3)

        # confirm columns are as expected
        if columns is not None:
            assert np.all(columns == data.columns)

        # confirm errors
        with pytest.raises(TypeError):
            func(1)

        with pytest.raises(TypeError):
            func([1, 2, 3])

        with pytest.raises(FileNotFoundError):
            func('notafile')



# -*- coding: utf-8 -*-
"""
Tests for abagen.cli module
"""

import os
from pkg_resources import resource_filename
import pytest

from abagen import __version__ as version
from abagen.cli import run


def test_run_get_parser(capsys, atlas, datadir):
    parser = run.get_parser()

    # since we have a positional argument this fails, hard
    with pytest.raises(SystemExit):
        parser.parse_args([])
    assert "following arguments are required: atlas" in capsys.readouterr().err

    # providing the positional succeeds!
    args = parser.parse_args([atlas['image']])
    assert os.path.normcase(args.atlas) == os.path.normcase(atlas['image'])

    # some data directories/files need to exist!
    with pytest.raises(SystemExit):
        parser.parse_args(['notanatlas.nii.gz'])
    assert "does not exist" in capsys.readouterr().err
    with pytest.raises(SystemExit):
        parser.parse_args(['--data-dir', 'notadir', atlas['image']])
    assert "does not exist" in capsys.readouterr().err
    with pytest.raises(SystemExit):
        parser.parse_args(['--atlas-info', 'notafile', atlas['image']])
    assert "does not exist" in capsys.readouterr().err

    # does version print correctly?
    with pytest.raises(SystemExit):
        parser.parse_args(['--version'])
    assert 'abagen {}'.format(version) == capsys.readouterr().out.strip()

    # arguments with invalid choices (probe_selection) raise errors
    with pytest.raises(SystemExit):
        parser.parse_args(['--probe-selection', 'notamethod', atlas['image']])
    assert "invalid choice: 'notamethod'" in capsys.readouterr().err
    with pytest.raises(SystemExit):
        parser.parse_args(['--gene-norm', 'notamethod', atlas['image']])
    assert "invalid choice: 'notamethod'" in capsys.readouterr().err
    with pytest.raises(SystemExit):
        parser.parse_args(['--sample-norm', 'notamethod', atlas['image']])
    assert "invalid choice: 'notamethod'" in capsys.readouterr().err

    # just test every option
    args = parser.parse_args([
        '-v', '-v', '-v',
        '--quiet',
        '--debug',
        '--atlas-info', atlas['info'],
        '--donors', '12876', '15496',
        '--data-dir', datadir,
        '--inexact',
        '--tol', '5',
        '--ibf-threshold', '0.6',
        '--region-agg', 'donors',
        '--agg-metric', 'median',
        '--probe-selection', 'average',
        '--lr_mirror',
        '--gene-norm', 'srs',
        '--sample-norm', 'srs',
        '--no-reannotated', '--no-corrected-mni',
        '--output-file', 'test.csv',
        '--save-counts', '--save-donors',
        atlas['image']
    ])


def test_run_main(capsys, atlas, datadir):
    outputfile = os.path.join(datadir, 'abagen_expression.csv')

    # check basic usage
    run.main([
        '--data-dir', datadir,
        '--donors', '12876', '15496',
        '--output-file', outputfile,
        atlas['image']
    ])
    assert os.path.exists(outputfile)

    # check that save donors/counts outputs desired files
    run.main([
        '--data-dir', datadir,
        '--donors', '12876', '15496',
        '--output-file', outputfile,
        '--save-donors', '--save-counts',
        atlas['image']
    ])
    for rep in ['_counts.csv', '_12876.csv', '_15496.csv']:
        assert os.path.exists(outputfile.replace('.csv', rep))

    # check stdout (BLARGH)
    run.main([
        '--data-dir', datadir,
        '--donors', '12876', '15496',
        '--output-file', outputfile,
        '--stdout',
        atlas['image']
    ])
    stdout = capsys.readouterr().out.replace('\r\n', '\n').replace('\r', '\n')
    with open(outputfile, 'r') as src:
        data = src.read().replace('\r\n', '\n').replace('\r', '\n')
        assert stdout == data


def test_exec_run_fail():
    executable = resource_filename('abagen', 'cli/run.py')

    # need to set this otherwise it won't fail
    __name__ = '__main__'  # noqa
    with pytest.raises(RuntimeError):
        with open(executable, 'r') as src:
            exec(src.read())



# -*- coding: utf-8 -*-
"""
Functions to make mouse gene queries and manipulations
"""

import re
import pandas as pd
from .utils import _coerce_inputs, _make_api_query


# available attributes of gene query
_GENE_ATTRIBUTES = [
    'acronym',
    'alias_tags',
    'chromosome_id',
    'ensembl_id',
    'entrez_id',
    'genomic_reference_update_id',
    'homologene_id',
    'id',
    'legacy_ensembl_gene_id',
    'name',
    'organism_id',
    'original_name',
    'original_symbol',
    'reference_genome_id',
    'sphinx_id',
    'version_status'
]


def available_gene_info():
    """ Lists available attributes for :func:`abagen.mouse.get_gene_info`
    """

    return _GENE_ATTRIBUTES.copy()


def get_gene_info(id=None, acronym=None, name=None, attributes=None,
                  verbose=False):
    """
    Queries Allen API for information about given gene

    One of `id`, `acronym`, or `name` must be provided.

    Parameters
    ----------
    id : int, optional
        Numerical gene ID
    acronym : str, optional
        Short-form gene acronym (case sensitive)
    name : str, optional
        Full gene name (case sensitive)
    attributes : str or list, optional
        Which attributes / information to obtain for the provided gene. See
        :func:`abagen.mouse.available_gene_info` for list of available
        attributes to request. If not specified all available attributes will
        be returned. Default: None
    verbose : bool, optional
        Whether to print status messages. Default: False

    Returns
    -------
    info : pandas.DataFrame
        If `attributes` is a str, returns an int or str depending on specified
        attribute. If `attributes` is a list, return a dict where keys are
        attributes and values are str or int.

    Raises
    ------
    ValueError
        The provided gene is invalid

    Examples
    --------
    Get gene ID and name corresponding to gene acronym 'Pdyn':
    >>> from abagen import mouse
    >>> mouse.get_gene_info(acronym='Pdyn', attributes=['id', 'name'])
                id          name
    acronym
    Pdyn     18376  prodynorphin

    You can also supply multiple genes to the query:
    >>> mouse.get_gene_info(acronym=['Ace', 'Cd99'], attributes=['id', 'name'])
                 id                                               name
    acronym
    Ace       11210  angiotensin I converting enzyme (peptidyl-dipe...
    Cd99     163028                                       CD99 antigen
    """

    criteria = [
        _coerce_inputs(id=id, acronym=acronym, name=name),
        'products[id$eq1]'
    ]
    provided = re.search(r'\[(\S+)\$', criteria[0]).group(1)

    # determine which attributes to request; if we don't have to request all
    # of them then we can speed up the API call
    if attributes is None:
        attributes = _GENE_ATTRIBUTES
    elif isinstance(attributes, str):
        attributes = [attributes]
    attributes = [a for a in attributes if a not in provided]

    for attr in attributes:
        if attr not in _GENE_ATTRIBUTES:
            raise ValueError('Provided attribute "{}" is invalid; please '
                             'check valid attributes with '
                             'abagen.mouse.available_gene_info().'
                             .format(attr))

    info = _make_api_query('Gene', criteria=criteria,
                           attributes=attributes + [provided], verbose=verbose)
    info = pd.DataFrame(info).set_index(provided)[attributes]

    return info.sort_index()



# -*- coding: utf-8 -*-

import json
from typing import Iterable
import urllib
from urllib.request import HTTPError, urlopen


def _coerce_inputs(id=None, acronym=None, name=None):
    """
    Checks that at least one of required inputs was supplied

    Parameters
    ----------
    id : int, optional
        Numerical ID
    acronym : str, optional
        Short-form acronym (case sensitive)
    name : str, optional
        Full name (case sensitive)

    Returns
    -------
    criteria : str
        Should be provided to `criteria` kwarg of :func:`_make_api_query`
    """

    # preferred order of inputs: id > acronym > name
    query, params = 'eq', [id, acronym, name]
    for criteria, param in zip(['id', 'acronym', 'name'], params):
        if param is not None:
            if isinstance(param, Iterable) and not isinstance(param, str):
                # if we're doing "in" queries with strings they must be quoted
                if criteria in ['acronym', 'name']:
                    param = ["'{}'".format(f) for f in param]
                query, param = 'in', ','.join([str(f) for f in param])
            break
    if param is None:
        params = ['id', 'acronym', 'name']
        raise TypeError('At least one of {} must be specified.'.format(params))

    return '[{}${}{}]'.format(criteria, query, param)


def _make_api_query(dtype, includes=None, criteria=None, attributes=None,
                    suffix=None, returns='msg', verbose=False):
    """
    """

    url = 'https://api.brain-map.org/api/v2/data/{}/query.json?'.format(dtype)

    params = [includes, criteria, attributes]
    for key, value in zip(['include', 'criteria', 'only'], params):
        if value is not None:
            if isinstance(value, list):
                value = ','.join(value)
            url += '{}={}&'.format(key, urllib.parse.quote_plus(value))

    if suffix is not None:
        url += suffix

    if verbose:
        print("Querying {}...".format(urllib.parse.unquote_plus(url)))
    response = urlopen(url)
    if response.status != 200:
        raise HTTPError('Failed to query API with code {}: {}'
                        .format(response.status, response.reason))

    info = json.loads(response.read().decode('utf-8'))

    if not info['success']:
        raise ValueError('Provided query {} is invalid. Please check '
                         'parameters and try again.'
                         .format(urllib.parse.unquote_plus(url)))
    elif info['total_rows'] == 0:
        raise ValueError('Provided query {} returned no results. Please '
                         'check parameters and try again.'
                         .format(urllib.parse.unquote_plus(url)))

    if returns is not None:
        info = info.get(returns, [])

    return info



__all__ = [
    'available_gene_info', 'get_gene_info', 'available_structure_info',
    'get_structure_info', 'get_structure_coordinates', 'fetch_allenref_genes',
    'fetch_allenref_structures', 'fetch_rubinov2015_structures',
    'get_unionization_from_gene', 'available_unionization_info'
]

from .gene import (get_gene_info, available_gene_info)
from .structure import (get_structure_info, available_structure_info,
                        get_structure_coordinates)
from .io import (fetch_allenref_genes, fetch_allenref_structures,
                 fetch_rubinov2015_structures)
from .mouse import (get_unionization_from_gene, available_unionization_info)



# -*- coding: utf-8 -*-
"""
Functions for annotating / selecting microarray probes
"""

import functools
import gzip
from io import StringIO
import itertools
import logging
from pkg_resources import resource_filename

import numpy as np
import pandas as pd
from scipy import stats as sstats

from . import datasets, io, utils

lgr = logging.getLogger('abagen')


def reannotate_probes(probes):
    """
    Replaces gene symbols in `probes` with reannotated data

    Uses annotations from [PR18]_ to replace probe annotations shipped with
    AHBA data. Any probes that were unable to be matched to a gene in
    reannotation procedure are not retained.

    Parameters
    ----------
    probes : str or pandas.DataFrame
        Probe file or loaded probe dataframe from Allen Brain Institute
        containing information on microarray probes

    Returns
    -------
    reannotated : pandas.DataFrame
        Provided probe information with updated gene symbols and Entrez IDs

    References
    ----------
    .. [PR18] Arnatkevic̆iūtė, A., Fulcher, B. D., & Fornito, A. (2019). A
       practical guide to linking brain-wide gene expression and neuroimaging
       data. NeuroImage, 189, 353-367.
    """

    lgr.info('Reannotating probes with information from Arnatkevic̆iūtė '
             'et al., 2019, NeuroImage')

    # load in reannotated probes
    reannot = resource_filename('abagen', 'data/reannotated.csv.gz')
    with gzip.open(reannot, 'r') as src:
        reannot = pd.read_csv(StringIO(src.read().decode('utf-8')))

    # merge reannotated with original, keeping only reannotated
    probes = io.read_probes(probes).reset_index()
    merged = pd.merge(reannot[['probe_name', 'gene_symbol', 'entrez_id']],
                      probes[['probe_name', 'probe_id']],
                      on='probe_name', how='left')

    # reset index as probe_id and sort
    reannotated = merged.set_index('probe_id') \
                        .sort_index() \
                        .dropna(subset=['entrez_id'])
    reannotated.loc[:, 'entrez_id'] = reannotated['entrez_id'].astype('int')

    return reannotated


def filter_probes(pacall, annotation, probes, threshold=0.5):
    """
    Performs intensity based filtering (IBF) of expression probes

    Uses binary indicator for expression levels in `pacall` to determine which
    probes have expression levels above background noise in `threshold` of
    samples across donors.

    Parameters
    ----------
    pacall : dict
        Dictionary where keys are donor IDs and values are filepaths to (or
        dataframes of) PACall.csv files from Allen Brain Institute
    annotation : dict
        Dictionary where keys are donor IDs and values are filepaths to (or
        dataframes of) SampleAnnot.csv files from Allen Brain Institute
    probes : str or pandas.DataFrame
        Filepath to Probes.csv or dataframe containing information on
        microarray probes that should be considered in filtering (probes not in
        this will be ignored)
    threshold : (0, 1) float, optional
        Threshold for filtering probes. Specifies the proportion of samples for
        which a given probe must have expression levels above background noise.
        Default: 0.5

    Returns
    -------
    filtered : pandas.DataFrame
        Dataframe containing information on probes that should be retained
        according to intensity-based filtering
    """

    threshold = np.clip(threshold, 0.0, 1.0)

    lgr.info(f'Filtering probes with intensity-based threshold of {threshold}')

    probes = io.read_probes(probes)
    signal, n_samp = np.zeros(len(probes), dtype=int), 0
    for donor, pa in pacall.items():
        annot = io.read_annotation(annotation[donor]).index
        data = io.read_pacall(pa).loc[probes.index, annot]
        n_samp += data.shape[-1]
        # sum binary expression indicator across samples for current subject
        signal += np.asarray(data.sum(axis=1))

    # calculate proportion of signal to noise for given probe across samples
    keep = (signal / n_samp) >= threshold

    lgr.info(f'{keep.sum()} probes survive intensity-based filtering')

    return probes[keep]


def _groupby_structure_id(microarray, annotation):
    """
    Averages samples in `microarray` having identical structure IDs

    Parameters
    ----------
    microarray : (P, S) pandas.DataFrame
        Dataframe should have `P` rows representing probes and `S` columns
        representing distinct samples, with values indicating microarray
        expression levels
    annotation : (S, A) pandas.DataFrame
        Annotation dataframe, obtained by loading a SampleAnnot.csv file from
        Allen Brain Institute

    Returns
    -------
    expression : (P, R) pandas.DataFrame
        Input `microarray` dataframe but with `S` samples averaged into `R`
        regions
    """

    sid = io.read_annotation(annotation)['structure_id']

    return io.read_microarray(microarray).groupby(sid, axis=1).mean()


def _groupby_and_apply(expression, probes, info, applyfunc):
    """
    Subsets `expression` based on most representative probe

    Parameters
    ----------
    expression : dict of (P, S) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `P`
        rows representing probes and `S` columns representing distinct samples
    probes : pandas.DataFrame
        Dataframe containing information on probes that should be considered in
        representative analysis. Generally, intensity-based-filtering (i.e.,
        `filter_probes()`) should have been used to reduce this list to only
        those probes with good expression signal
    info : pandas.DataFrame
        Dataframe containing information on probe expression information. Index
        should be unique probe IDs and must have at least 'gene_symbol' column
    applyfunc : callable
        Function used to select representative probe ID from those indexing
        the same gene. Must accept a pandas dataframe as input and return a
        string (i.e., the chosen probe ID)

    Returns
    -------
    representative : dict of (S, G) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `S`
        rows representing distinct samples and `G` columns representing unique
        genes
    """

    # group probes by gene and get probe corresponding to relevant feature
    retained = info.groupby('gene_symbol').apply(applyfunc).dropna().squeeze()
    probes = probes.loc[sorted(retained)].sort_values('gene_symbol')

    # subset expression dataframes to retain only desired probes
    representative = {
        d: e.loc[probes.index].T
        for d, e in utils.check_dict(expression).items()
    }

    return representative


def _diff_stability(expression, probes, annotation, *args, **kwargs):
    """
    Picks one probe to represent `expression` data for each gene in `probes`

    If there are multiple probes with expression data for the same gene, this
    function will calculate the similarity of each probes' expression across
    donors and select the probe with the most consistent pattern of regional
    variation (i.e., "differential stability" or DS). Regions are defined by
    the "structure_id" column in `annotation`; similarity is calculated by the
    Spearman correlation coefficient

    Parameters
    ----------
    expression : dict of (P, S) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `P`
        rows representing probes and `S` columns representing distinct samples
    probes : pandas.DataFrame
        Dataframe containing information on probes that should be considered in
        representative analysis. Generally, intensity-based-filtering (i.e.,
        `filter_probes()`) should have been used to reduce this list to only
        those probes with good expression signal
    annotation : list of str
        List of filepaths to annotation files from Allen Brain Institute (i.e.,
        as obtained by calling :func:`abagen.fetch_microarray` and accessing
        the `annotation` attribute on the resulting object).

    Returns
    -------
    representative : dict of (S, G) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `S`
        rows representing distinct samples and `G` columns representing unique
        genes
    """

    # confirm inputs are expected dictionaries
    expression = utils.check_dict(expression)
    annotation = utils.check_dict(annotation)

    # collapse (i.e., average) expression across AHBA anatomical regions
    region_exp = [
        _groupby_structure_id(microarray, annotation[donor])
        for donor, microarray in expression.items()
    ]

    # get correlation of probe expression across samples for all donor pairs
    probe_exp = np.zeros((len(probes), sum(range(len(expression)))))
    for n, (exp1, exp2) in enumerate(itertools.combinations(region_exp, 2)):
        # samples that current donor pair have in common
        samples = np.intersect1d(exp1.columns, exp2.columns)

        # the ranking process can take a few seconds on each loop
        # unfortunately, we have to do it each time because `samples` changes
        # based on which anatomical regions the two subjects have in common
        exp1 = exp1.loc[:, samples].T.rank()
        exp2 = exp2.loc[:, samples].T.rank()

        probe_exp[:, n] = utils.efficient_corr(exp1, exp2)

    info = pd.DataFrame(dict(gene_symbol=np.asarray(probes.gene_symbol),
                             diff_stability=probe_exp.mean(axis=1)),
                        index=probes.index)
    applyfunc = functools.partial(_max_idx, column='diff_stability')

    return _groupby_and_apply(expression, probes, info, applyfunc)


def _rnaseq(expression, probes, annotation, *args, **kwargs):
    """

    Parameters
    ----------
    expression : dict of (P, S) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `P`
        rows representing probes and `S` columns representing distinct samples
    probes : pandas.DataFrame
        Dataframe containing information on probes that should be considered in
        representative analysis. Generally, intensity-based-filtering (i.e.,
        `filter_probes()`) should have been used to reduce this list to only
        those probes with good expression signal
    annotation : list of str
        List of filepaths to annotation files from Allen Brain Institute (i.e.,
        as obtained by calling :func:`abagen.fetch_microarray` and accessing
        the `annotation` attribute on the resulting object).

    Returns
    -------
    representative : dict of (S, G) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `S`
        rows representing distinct samples and `G` columns representing unique
        genes
    """
    # confirm inputs are expected dictionaries
    expression = utils.check_dict(expression)
    annotation = utils.check_dict(annotation)

    # fetch RNAseq data
    rnaseq = datasets.fetch_rnaseq(donors=expression.keys())

    probe_exp = np.ones((len(probes), len(rnaseq))) * np.nan
    for n, (donor, data) in enumerate(rnaseq.items()):
        # collapse (i.e., average) data across AHAB  anatomical regions
        micro = _groupby_structure_id(expression[donor], annotation[donor])
        rna = _groupby_structure_id(io.read_tpm(data['tpm']),
                                    data['annotation'])

        # get rid of "constant" RNAseq genes
        rna = rna[np.logical_not(np.isclose(rna.std(axis=1, ddof=1), 0))]

        # get matching genes + strcutres between microarray + RNAseq
        regions = np.intersect1d(micro.columns, rna.columns)
        mask = np.isin(np.asarray(probes.gene_symbol),
                       np.intersect1d(probes.gene_symbol, rna.index))
        genes = np.asarray(probes.loc[mask, 'gene_symbol'])
        micro, rna = micro.loc[mask, regions].T, rna.loc[genes, regions].T

        # correlate expression values across regions for each gene
        probe_exp[mask, n] = utils.efficient_corr(micro.rank(), rna.rank())

    mask = np.sum(np.isnan(probe_exp), axis=1) < len(rnaseq)
    info = pd.DataFrame(dict(gene_symbol=np.asarray(probes[mask].gene_symbol),
                             rna_corr=np.nanmean(probe_exp[mask], axis=1)),
                        index=probes.index[mask])
    applyfunc = functools.partial(_max_idx, column='rna_corr')

    return _groupby_and_apply(expression, probes, info, applyfunc)


def _max_idx(df, column=0):
    """
    Returns probe ID with max index in `df`

    Parameters
    ----------
    df : (P, 1) pandas.DataFrame
        Dataframe with `P` rows indicating distinct probes and one column
        containing summary statistic of probe expression
    column : str, optional
        Column name from which to extract the max index. If not specified uses
        the first numerical column.

    Returns
    -------
    probe_id : str
        ID of probe selected as representative for given gene
    """

    return df.idxmax()[column]


def _max_loading(df):
    """
    Returns probe ID with max loading along first principal component of `df`

    Parameters
    ----------
    df : (P, S) pandas.DataFrame
        Dataframe with `P` rows indicating distinct probe expression values
        across `S` samples

    Returns
    -------
    probe_id : str
        ID of probe selected as representative for given gene
    """

    if len(df) == 1:
        return df.index[0]

    data = np.asarray(df)
    data = data - data.mean(axis=0, keepdims=True)

    # svd() is faster than eig() here because we don't need to construct the
    # covariance matrix
    u, s, v = np.linalg.svd(data, full_matrices=False)

    # use sign flip based on right singular vectors (as we would with eig())
    v *= np.sign(v[range(len(v)), np.argmax(np.abs(v), axis=1)])[:, np.newaxis]

    return df.index[(data @ v.T)[:, 0].argmax()]


def _correlate(df, method):
    """
    Returns probe ID with max avg correlation (>2 probes) or `method` (<2)

    Parameters
    ----------
    df : (P, S) pandas.DataFrame
        Dataframe with `P` rows indicating distinct probe expression values
        across `S` samples
    method : {'variance', 'intensity'}
        Method for selecting representative probe when only two probes index
        the same gene (>2 probes uses correlation method)

    Returns
    -------
    probe_id : str
        ID of probe selected as representative for given gene
    """

    data = np.asarray(df)

    if len(data) > 2:
        xmax = np.mean((np.corrcoef(data) + 1) / 2, axis=1)
    elif len(data) == 2:
        if method == 'variance':
            xmax = np.std(data, axis=1)
        elif method == 'intensity':
            xmax = np.mean(data, axis=1)
    else:
        return df.index[0]

    return df.index[xmax.argmax()]


def _average(expression, probes, *args, **kwargs):
    """
    Averages expression data for probes representing the same gene

    Parameters
    ----------
    expression : dict of (P, S) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `P`
        rows representing probes and `S` columns representing distinct samples
    probes : pandas.DataFrame
        Dataframe containing information on microarray probes that should be
        considered in representative analysis. Generally intensity-based
        filtering (i.e., via :func:`filter_probes()`) should have been used to
        reduce this list to only those probes with good expression signal

    Returns
    -------
    representative : dict of (S, G) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `S`
        rows representing distinct samples and `G` columns representing unique
        genes
    """

    def _avg(df):
        return df.join(probes['gene_symbol'], on='probe_id') \
                 .groupby('gene_symbol') \
                 .mean().T

    return {d: _avg(exp) for d, exp in utils.check_dict(expression).items()}


def _collapse(expression, probes, *args, method='max_variance', **kwargs):
    """
    Selects one representative probe per gene using provided `method`

    Parameters
    ----------
    expression : dict of (P, S) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `P`
        rows representing probes and `S` columns representing distinct samples
    probes : pandas.DataFrame
        Dataframe containing information on microarray probes that should be
        considered in representative analysis. Generally intensity-based
        filtering (i.e., via :func:`filter_probes()`) should have been used to
        reduce this list to only those probes with good expression signal
    method : str, optional
        Method by which to select represenative probes for each gene. Must be
        one of ['max_variance', 'max_intensity', 'pc_loading', 'corr_variance',
        'corr_intensity']

    Returns
    -------
    representative : dict of (S, G) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `S`
        rows representing distinct samples and `G` columns representing unique
        genes
    """

    # concatenate all donors into giant probe x sample expression dataframe
    expression = utils.check_dict(expression)
    probe_exp = pd.concat(expression.values(), axis=1)

    # determine aggregation function based on provided method; also reduce
    # probe expression if required (i.e., max_variance, max_intensity)
    if method == 'max_variance':
        probe_exp = pd.DataFrame(probe_exp.std(axis=1), columns=[method])
        agg = functools.partial(_max_idx, column=method)
    elif method == 'max_intensity':
        probe_exp = pd.DataFrame(probe_exp.mean(axis=1), columns=[method])
        # probe_exp.name = method
        agg = functools.partial(_max_idx, column=method)
    elif method == 'pc_loading':
        agg = _max_loading
    elif method in ['corr_variance', 'corr_intensity']:
        agg = functools.partial(_correlate, method=method[5:])
    else:
        raise ValueError(f'Provided method {method} is invalid. Please check '
                         'inputs and try again.')

    info = pd.merge(probes[['gene_symbol']], probe_exp, on='probe_id')
    return _groupby_and_apply(expression, probes, info, agg)


_max_variance = functools.partial(_collapse, method='max_variance')
_max_intensity = functools.partial(_collapse, method='max_intensity')
_pc_loading = functools.partial(_collapse, method='pc_loading')
_corr_variance = functools.partial(_collapse, method='corr_variance')
_corr_intensity = functools.partial(_collapse, method='corr_intensity')


SELECTION_METHODS = dict(
    mean=_average,
    average=_average,
    max_intensity=_max_intensity,
    max_variance=_max_variance,
    pc_loading=_pc_loading,
    corr_variance=_corr_variance,
    corr_intensity=_corr_intensity,
    diff_stability=_diff_stability,
    rnaseq=_rnaseq
)
AGG_METHODS = [  # can only be used with `donor_probes='aggregate'`
    'mean', 'average', 'diff_stability', 'rnaseq'
]
COLLAPSE_METHODS = [  # methods that don't SELECT but COLLAPSE ACROSS probes
    'mean', 'average'
]


def collapse_probes(microarray, annotation, probes, method='diff_stability',
                    donor_probes='aggregate'):
    """
    Reduces `microarray` to a sample x gene expression dataframe

    Using provided `method`, reduces `microarray` expression data by either (1)
    selecting a representative probe amongst all probes indexing the same gene,
    or (2) collapsing across all probes indexing the same gene. See Notes for
    more information on different methods available.

    Parameters
    ----------
    microarray : dict of str or pandas.DataFrame
        Dictionary where keys are donor IDs and values are filepaths to (or
        dataframes of) MicroarrayExpression.csv files from Allen Brain
        Institute
    annotation : dict of str or pandas.DataFrame
        Dictionary where keys are donor IDs and values are filepaths to (or
        dataframes of) SampleAnnot.csv files from Allen Brain Institute. Only
        used if `method='diff_stability'`
    probes : str or pandas.DataFrame
        Filepath to Probes.csv or dataframe containing information on
        microarray probes that should be considered in representative analysis.
        Generally intensity-based filtering (i.e., via :func:`filter_probes()`)
        should have been used to reduce this list to only those probes with
        good expression signal
    method : str, optional
        Selection method for subsetting (or collapsing across) probes from the
        same gene. Must be one of 'average', 'max_intensity', 'max_variance',
        'pc_loading', 'corr_intensity', 'corr_variance', 'diff_stability', or
        'rnaseq'; see Notes for more information. Default: 'diff_stability'
    donor_probes : str, optional
        Whether specified `probe_selection` method should be performed with
        microarray data from all donors ('aggregate'), independently for each
        donor ('independent'), or based on the most common selected probe
        across donors ('common'). Not all combinations of `probe_selection`
        and `donor_probes` methods are viable. Default: 'aggregate'

    Returns
    -------
    expression : dict of (S, G) pandas.DataFrame
        Dictionary where keys are donor IDs and values are dataframes with `S`
        rows representing distinct samples and `G` columns representing unique
        genes. Entries of dataframe indicate microarray expression levels for
        each combination of sample + gene. Columns will be identical across all
        dataframes, but `S` will vary by donor.

    Notes
    -----
    The following methods can be used for collapsing across probes when
    multiple probes are available for the same gene.

    1. ``method='average'``

    Uses the average of expression data across probes indexing the same gene as
    in [PR3]_, [PR5]_, [PR9]_, [PR10]_, [PR15]_, [PR16]_, and [PR17]_. Using
    `method='mean'` will do the same thing.

    2. ``method='max_intensity'``

    Selects probe with maximum average expression across samples from all
    donors as in [PR14]_.

    3. ``method='max_variance'``

    Selects probe with maximum variance in expression across samples from all
    donors as in [PR12]_.

    4. ``method='pc_loading'``

    Selects probe with maximum loading on first principal component of
    decomposition performed across samples from all donors as in [PR13]_.

    5. ``method='corr_intensity'``

    Selects probe with maximum correlation to other probes from same gene when
    >2 probes exist; otherwise, uses same procedure as `method=max_intensity`.
    Used in [PR1]_ and [PR11]_.

    6. ``method='corr_variance'``

    Selects probe with maximum correlation to other probes from same gene when
    >2 probes exist; otherwise, uses same procedure as `method=max_variance`.
    Used in [PR2]_, [PR4]_, and [PR6]_.

    7. ``method='diff_stability'``

    Selects probe with the most consistent pattern of regional variation across
    donors (i.e., highest average correlation across brain regions between all
    pairs of donors) as in [PR7]_ and [PR8]_.

    8. ``method='rnaseq'``

    Selects probes with most consistent pattern of regional variation to RNAseq
    data (across the two donors with RNAseq data).

    References
    ----------
    .. [PR1] Anderson, K. M., Krienen, F. M., Choi, E. Y., Reinen, J. M., Yeo,
       B. T., & Holmes, A. J. (2018). Gene expression links functional networks
       across cortex and striatum. Nature Communications, 9(1), 1428.

    .. [PR2] Burt, J. B., Demirtaş, M., Eckner, W. J., Navejar, N. M., Ji, J.
       L., Martin, W. J., ... & Murray, J. D. (2018). Hierarchy of
       transcriptomic specialization across human cortex captured by structural
       neuroimaging topography. Nature Neuroscience, 21(9), 1251.

    .. [PR3] Eising, E., Huisman, S. M., Mahfouz, A., Vijfhuizen, L. S.,
       Anttila, V., Winsvold, B. S., ... & Boomsma, D. I. (2016). Gene
       co-expression analysis identifies brain regions and cell types involved
       in migraine pathophysiology: a GWAS-based study using the Allen Human
       Brain Atlas. Human Genetics, 135(4), 425-439.

    .. [PR4] Forest, M., Iturria‐Medina, Y., Goldman, J. S., Kleinman, C. L.,
       Lovato, A., Oros Klein, K., ... & Greenwood, C. M. (2017). Gene networks
       show associations with seed region connectivity. Human Brain Mapping,
       38(6), 3126-3140.

    .. [PR5] French, L., & Paus, T. (2015). A FreeSurfer view of the cortical
       transcriptome generated from the Allen Human Brain Atlas. Frontiers in
       Neuroscience, 9, 323.

    .. [PR6] Hawrylycz, M. J., Lein, E. S., Guillozet-Bongaarts, A. L., Shen,
       E. H., Ng, L., Miller, J. A., ... & Abajian, C. (2012). An anatomically
       comprehensive atlas of the adult human brain transcriptome. Nature,
       489(7416), 391.

    .. [PR7] Hawrylycz, M., Miller, J. A., Menon, V., Feng, D., Dolbeare, T.,
       Guillozet-Bongaarts, A. L., ... & Glasser, M. F. (2015). Canonical
       genetic signatures of the adult human brain. Nature Neuroscience,
       18(12), 1832.

    .. [PR8] Kirsch, L., & Chechik, G. (2016). On expression patterns and
       developmental origin of human brain regions. PLoS Computational Biology,
       12(8), e1005064.

    .. [PR9] Krienen, F. M., Yeo, B. T., Ge, T., Buckner, R. L., & Sherwood, C.
       C. (2016). Transcriptional profiles of supragranular-enriched genes
       associate with corticocortical network architecture in the human brain.
       Proceedings of the National Academy of Sciences, 113(4), E469-E478.

    .. [PR10] McColgan, P., Gregory, S., Seunarine, K. K., Razi, A., Papoutsi,
       M., Johnson, E., ... & Scahill, R. I. (2018). Brain regions showing
       white matter loss in Huntington’s disease are enriched for synaptic and
       metabolic genes. Biological Psychiatry, 83(5), 456-465.

    .. [PR11] Myers, E. M., Bartlett, C. W., Machiraju, R., & Bohland, J. W.
       (2015). An integrative analysis of regional gene expression profiles in
       the human brain. Methods, 73, 54-70.S

    .. [PR12] Negi, S. K., & Guda, C. (2017). Global gene expression profiling
       of healthy human brain and its application in studying neurological
       disorders. Scientific Reports, 7(1), 897.

    .. [PR13] Parkes, L., Fulcher, B. D., Yücel, M., & Fornito, A. (2017).
       Transcriptional signatures of connectomic subregions of the human
       striatum. Genes, Brain and Behavior, 16(7), 647-663.

    .. [PR14] Romero-Garcia, R., Whitaker, K. J., Váša, F., Seidlitz, J.,
       Shinn, M., Fonagy, P., ... & Vértes, P. E. (2018). Structural covariance
       networks are coupled to expression of genes enriched in supragranular
       layers of the human cortex. NeuroImage, 171, 256-267.

    .. [PR15] Tan, P. P. C., French, L., & Pavlidis, P. (2013). Neuron-enriched
       gene expression patterns are regionally anti-correlated with
       oligodendrocyte-enriched patterns in the adult mouse and human brain.
       Frontiers in Neuroscience, 7, 5.

    .. [PR16] Vértes, P. E., Rittman, T., Whitaker, K. J., Romero-Garcia, R.,
       Váša, F., Kitzbichler, M. G., ... & Goodyer, I. M. (2016). Gene
       transcription profiles associated with inter-modular hubs and connection
       distance in human functional magnetic resonance imaging networks.
       Philosophical Transactions of the Royal Society B: Biological Sciences,
       371(1705), 20150362.

    .. [PR17] Whitaker, K. J., Vértes, P. E., Romero-Garcia, R., Váša, F.,
       Moutoussis, M., Prabhu, G., ... & Tait, R. (2016). Adolescence is
       associated with genomically patterned consolidation of the hubs of the
       human brain connectome. Proceedings of the National Academy of Sciences,
       113(32), 9105-9110.
    """

    try:
        collfunc = SELECTION_METHODS[method]
    except KeyError:
        raise ValueError(f'Provided `method` "{method}" is invalid; must be '
                         f'one of {list(SELECTION_METHODS)}')

    valid_probes = ['aggregate', 'common', 'independent']
    if donor_probes not in valid_probes:
        raise ValueError(f'Provided `donor_probes` "{donor_probes}" is '
                         f'invalid; must be one of {valid_probes}')

    lgr.info(f'Reducing probes indexing same gene with method: {method}')
    # subset microarray data for pre-selected probes + samples
    # this will also left/right mirror samples, if previously requested
    probes = io.read_probes(probes)
    microarray = utils.check_dict(microarray)
    annotation = utils.check_dict(annotation)
    for donor, micro in microarray.items():
        samp = io.read_annotation(annotation[donor]).index
        microarray[donor] = io.read_microarray(micro).loc[probes.index, samp]

    # now, "collect" the probes based on the provided `method`
    if method in AGG_METHODS or donor_probes == 'aggregate':
        # perform the collection function for all donors, together
        microarray = collfunc(microarray, probes, annotation)
    elif donor_probes == 'independent':
        # perform the collection function for each donor separately
        for donor in microarray:
            microarray.update(collfunc({donor: microarray[donor]}, probes,
                                       {donor: annotation[donor]}))
    elif donor_probes == 'common':
        # perform collection function for each donor separately and retain ONLY
        # the chose probe IDs
        probe_ids = [
            collfunc(
                {donor: microarray[donor]}, probes, {donor: annotation[donor]}
            )[donor].columns
            for donor in microarray
        ]
        # find the mode of the probe IDs chosen across donors and then subset
        # those probes from the original microarray dataframes for all donors
        probe_ids = np.squeeze(sstats.mode(probe_ids, axis=0)[0])
        for donor in microarray:
            microarray[donor] = microarray[donor].loc[probe_ids].T

    # convert probe IDs as column names to gene symbols
    if method not in COLLAPSE_METHODS:
        for donor, micro in microarray.items():
            symbols = probes.loc[micro.columns, 'gene_symbol']
            micro = micro.set_axis(symbols, axis=1, inplace=False)
            microarray[donor] = micro.sort_index(axis=1)

    n_genes = utils.first_entry(microarray).shape[-1]
    lgr.info(f'{n_genes} genes remain after probe filtering + selection')

    return microarray



# -*- coding: utf-8 -*-
"""
Tests for abagen.datasets.utils module
"""

import os
import shutil

from abagen.datasets import utils


def test_get_dataset_dir(datadir):
    old = os.environ.pop('ABAGEN_DATA', None)

    # check that data dir defaults to $HOME/abagen-data assuming no env var
    expected_base = os.path.join(os.path.expanduser('~'), 'abagen-data')
    data_dir = utils._get_dataset_dir('test', verbose=0)
    assert data_dir == os.path.join(expected_base, 'test')
    assert os.path.isdir(data_dir) and os.path.exists(data_dir)
    shutil.rmtree(data_dir)

    # if env var is set, we should default to that
    expected_base = os.path.join(datadir, 'test-abagen-data')
    os.environ['ABAGEN_DATA'] = expected_base
    data_dir = utils._get_dataset_dir('test', verbose=0)
    assert data_dir == os.path.join(expected_base, 'test')
    assert os.path.isdir(data_dir) and os.path.exists(data_dir)
    shutil.rmtree(expected_base)

    # test explicitly setting the data_dir
    expected_dir = datadir
    data_dir = utils._get_dataset_dir('test', data_dir=expected_dir,
                                      verbose=0)
    assert data_dir == os.path.join(expected_dir, 'test')
    assert os.path.isdir(data_dir) and os.path.exists(data_dir)
    # test that providing the returned data_dir gets us the same thing
    data_dir2 = utils._get_dataset_dir('test', data_dir=data_dir,
                                       verbose=0)
    assert data_dir == data_dir2
    shutil.rmtree(data_dir)

    if old is not None:
        os.environ['ABAGEN_DATA'] = old



# -*- coding: utf-8 -*-

import argparse
import logging
import os
from pathlib import Path
import sys

lgr = logging.getLogger('abagen')


def _resolve_path(path):
    """ Helper function for get_parser() to resolve paths
    """

    if path is not None:
        try:
            return str(Path(path).expanduser().resolve())
        except FileNotFoundError:
            return os.path.abspath(os.path.expanduser(path))


def _resolve_none(inp):
    """ Helper function to allow 'None' as input from argparse
    """

    if inp == "None":
        return
    return inp


class CheckExists(argparse.Action):
    """ Helper class to check that provided paths exist
    """
    def __call__(self, parser, namespace, values, option_string=None):
        values = self.type(values)
        if not os.path.exists(_resolve_path(values)):
            parser.error('Provided value for {} does not exist: {}'
                         .format(option_string, values))
        setattr(namespace, self.dest, values)


def get_parser():
    """ Gets command-line arguments for primary get_expression_data workflow
    """

    from .. import __version__
    from ..correct import NORMALIZATION_METHODS
    from ..probes_ import SELECTION_METHODS

    verstr = 'abagen {}'.format(__version__)
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
Assigns microarray expression data to ROIs defined in the specific `atlas`

This command aims to provide a workflow for generating pre-processed microarray
expression data from the Allen Human Brain Atlas for arbitrary atlas
designations. First, some basic filtering of genetic probes is performed,
including:

    1. Intensity-based filtering of microarray probes to remove probes that do
       not exceed a certain level of background noise (specified via the
       `ibf_threshold` parameter), and
    2. Selection of a single, representative probe (or collapsing across
       probes) for each gene, specified via the `probe_selection` parameter.

Tissue samples are then matched to parcels in the defined `atlas` for each
donor. If `atlas_info` is provided then this matching is constrained by both
hemisphere and tissue class designation (e.g., cortical samples from the left
hemisphere are only matched to ROIs in the left cortex, subcortical samples
from the right hemisphere are only matched to ROIs in the left subcortex); see
the `atlas_info` parameter description for more information.

Matching of microarray samples to parcels in `atlas` is done via a multi-step
process:

    1. Determine if the sample falls directly within a parcel,
    2. Check to see if there are nearby parcels by slowly expanding the search
       space to include nearby voxels, up to a specified distance (specified
       via the `tolerance` parameter),
    3. If there are multiple nearby parcels, the sample is assigned to the
       closest parcel, as determined by the parcel centroid.

If at any step a sample can be assigned to a parcel the matching process is
terminated. If multiple sample are assigned to the same parcel they are
aggregated with the metric specified via the `agg_metric` parameter. More
control over the sample matching can be obtained by setting the `inexact`
parameter; see the parameter description for more information.

Once all samples have been matched to parcels for all supplied donors, the
microarray expression data are optionally normalized via the provided
`sample_norm` and `gene_norm` functions before being combined within parcels
and across donors via the supplied `agg_metric`.
"""
    )

    parser.add_argument('atlas', action=CheckExists, type=_resolve_path,
                        help='An image in MNI space, where each parcel in the '
                             'image is identified by a unique integer ID.')

    # because I like consistency in capitalization and punctuation...
    for act in parser._actions:
        if isinstance(act, argparse._HelpAction):
            act.help = act.help.capitalize() + '.'
            break

    parser.add_argument('--version', action='version', version=verstr,
                        help='Show program\'s version number and exit.')
    parser.add_argument('-v', '--verbose', action='count', default=1,
                        help='Increase verbosity of status messages to '
                             'display during workflow.')
    parser.add_argument('-q', '--quiet', action='store_true', default=False,
                        help='Suppress all status messages during workflow.')
    parser.add_argument('--debug', action='store_true', help=argparse.SUPPRESS)

    a_data = parser.add_argument_group('Options to specify information about '
                                       'the atlas used')
    a_data.add_argument('--atlas_info', '--atlas-info', action=CheckExists,
                        type=_resolve_path, default=None, metavar='PATH',
                        help='Filepath to CSV files containing information '
                             'about `atlas`. The CSV file must have at least '
                             'columns ["id", "hemisphere", "structure"] which '
                             'contain information mapping the atlas IDs to '
                             'hemispheres (i.e, "L", "R") and broad '
                             'structural groups (i.e., "cortex", "subcortex", '
                             '"cerebellum", "brainstem").')

    g_data = parser.add_argument_group('Options to specify which AHBA data to '
                                       'use during processing')
    g_data.add_argument('--donors', action='store', nargs='+',
                        default='all', metavar='DONOR_ID',
                        help='List of donors to use as sources of expression '
                             'data. Specified IDs can be either donor numbers '
                             '(i.e., 9861, 10021) or UIDs (i.e., H0351.2001). '
                             'Can specify "all" to use all available donors. '
                             'Default: "all"')
    g_data.add_argument('--data_dir', '--data-dir', action=CheckExists,
                        type=_resolve_path, metavar='PATH',
                        help='Directory where expression data should be '
                             'downloaded to (if it does not already exist) / '
                             'loaded from. If not specified this will check '
                             'the environmental variable $ABAGEN_DATA, the '
                             '$HOME/abagen-data directory, and the current '
                             'working directory. If data does not already '
                             'exist at one of those locations then it will be '
                             'downloaded to the first of these location that '
                             'exists and for which write access is enabled.')

    w_data = parser.add_argument_group('Options to specify processing options')
    w_data.add_argument('--inexact', dest='exact', action='store_false',
                        default=True,
                        help='Whether to use inexact matching of donor tissue '
                             'samples to parcels in `atlas`. By default, the '
                             'workflow will match tissue samples to parcels '
                             'within `tolerance` mm of the sample; any '
                             'samples that are beyond `tolerance` mm of a '
                             'parcel will be discarded, which may result in '
                             'some parcels having no assigned sample / '
                             'expression data. If --inexact, the matching '
                             'procedure will be performed and followed by a '
                             'check for parcels with no assigned samples; any '
                             'such parcels will be matched to the nearest '
                             'sample (defined as the sample with the closest '
                             'Euclidean distance to the parcel centroid). '
                             'Default: False')
    w_data.add_argument('--tol', '--tolerance', dest='tolerance',
                        action='store', type=float, default=2,
                        help='Distance (in mm) that a sample can be from a '
                             'parcel for it to be matched to that parcel. '
                             'This is only considered if the sample is not '
                             'directly within a parcel. Default: 2')
    w_data.add_argument('--ibf_threshold', '--ibf-threshold', action='store',
                        default=0.5, metavar='THRESHOLD',
                        help='Threshold for intensity-based filtering of '
                             'probes. This number should specify the ratio of '
                             'samples, across all supplied donors, for which '
                             'a probe must have signal above background noise '
                             'in order to be retained. Default: 0.5')
    w_data.add_argument('--region_agg', '--region-agg', action='store',
                        default='donors', metavar='METHOD',
                        choices=['donors', 'samples'],
                        help='When multiple samples are identified as '
                             'belonging to a region in `atlas` this '
                             'determines how they are aggegated. If '
                             '\'samples\', expression data from all samples '
                             'for all donors assigned to a given region are '
                             'combined. If \'donors\', expression values for '
                             'all samples assigned to a given region are '
                             'combined independently for each donor before '
                             'being combined across donors. See `agg_metric` '
                             'for mechanism by which samples are combined. '
                             'Default: \'donors\'')
    w_data.add_argument('--agg_metric', '--agg-metric', action='store',
                        default='mean', metavar='METHOD',
                        choices=['mean', 'median'],
                        help='Mechanism by which to (1) reduce expression '
                             'data of multiple samples in the same `atlas` '
                             'region, and (2) reduce donor-level expression '
                             'data into a single "group" expression '
                             'dataframe. Must be one of {"mean", "median"}. '
                             'Default: "mean"')
    w_data.add_argument('--probe_selection', '--probe-selection',
                        action='store', default='diff_stability',
                        metavar='METHOD', choices=sorted(SELECTION_METHODS),
                        help='Selection method for subsetting (or collapsing '
                             'across) probes that index the same gene. Must '
                             'be one of {"average", "mean", "max_intensity", '
                             '"max_variance", "pc_loading", "corr_variance", '
                             '"corr_intensity", "diff_stability", "rnaseq"}. '
                             'Default: "diff_stability"')
    w_data.add_argument('--lr_mirror', '--lr-mirror', action='store_true',
                        help='Whether to mirror microarray expression samples '
                             'across hemispheres to increase spatial '
                             'coverage. This will duplicate samples across '
                             'both hemispheres (i.e., L->R and R->L), '
                             'approximately doubling the number of available '
                             'samples. Default: False (i.e., no mirroring)')
    w_data.add_argument('--gene_norm', '--gene-norm', action='store',
                        default='srs', metavar='METHOD', type=_resolve_none,
                        choices=sorted(NORMALIZATION_METHODS) + ['None', None],
                        help='Method by which to normalize microarray '
                             'expression values for each donor prior to '
                             'collapsing across donors. Expression values are '
                             'normalized separately for each gene for each '
                             'donor across all expression samples. If None is '
                             'specified then no normalization is performed. '
                             'Default: "srs"')
    w_data.add_argument('--sample_norm', '--sample-norm', action='store',
                        default='srs', metavar='METHOD', type=_resolve_none,
                        choices=sorted(NORMALIZATION_METHODS) + ['None', None],
                        help='Method by which to normalize microarray '
                             'expression values for each sample prior to '
                             'collapsing into regions in `atlas`. Expression '
                             'values are normalized separately for each '
                             'sample and donor across genes. If None is '
                             'specified then no normalization is performed. '
                             'Default: "srs"')

    p_data = parser.add_argument_group('Options to modify the AHBA data used')
    p_data.add_argument('--no-reannotated', '--no_reannotated',
                        dest='reannotated', action='store_false', default=True,
                        help='Whether to use the original probe information '
                             'from the AHBA dataset instead of the '
                             'reannotated probe information from '
                             'Arnatkevic̆iūtė et al., 2019. Using reannotated '
                             'probe information discards probes that could '
                             'not be reliably matched to genes. Default: '
                             'False (i.e., use reannotations)')
    p_data.add_argument('--no-corrected-mni', '--no_corrected_mni',
                        dest='corrected_mni', action='store_false',
                        default=True,
                        help='Whether to use the original MNI coordinates '
                             'provided with the AHBA data instead of the '
                             '"corrected" MNI coordinates shipped with the '
                             '`alleninf` package when matching tissue samples '
                             'to anatomical regions. Default: False (i.e., '
                             'use corrected coordinates)')

    o_data = parser.add_argument_group('Options to modify how data are output')
    o_data.add_argument('--stdout', action='store_true',
                        help='Generated region x gene dataframes will be '
                             'printed to stdout for piping to other things. '
                             'You should REALLY consider just using --output-'
                             'file instead and working with the generated '
                             'CSV file(s). Incompatible with `--save-counts` '
                             'and `--save-donors` (i.e., this will override '
                             'those options). Default: False')
    o_data.add_argument('--output-file', '--output_file', action='store',
                        type=_resolve_path, metavar='PATH',
                        default='abagen_expression.csv',
                        help='Path to desired output file. The generated '
                             'region x gene dataframe will be saved here. '
                             'Default: $PWD/abagen_expression.csv')
    o_data.add_argument('--save-counts', '--save_counts', action='store_true',
                        help='Whether to save dataframe containing number of '
                             'samples from each donor that were assigned '
                             'to each region in `atlas`. If specified, will '
                             'be saved to the path specified by '
                             '`output-file`, appending "counts" to the end of '
                             'the filename. Default: False')
    o_data.add_argument('--save-donors', '--save_donors', action='store_true',
                        help='Whether to save donor-level expression '
                             'dataframes instead of aggregating expression '
                             'across donors with provided `agg_metric`. If '
                             'specified, dataframes will be saved to path '
                             'specified by `output-file`, appending donor IDs '
                             'to the end of the filename. Default: False')

    return parser


def main(args=None):
    """ Runs primary get_expression_data workflow
    """

    from ..allen import get_expression_data
    from ..datasets import WELL_KNOWN_IDS as donors

    opts = get_parser().parse_args(args)

    # quiet overrides any verbosity setting
    if opts.quiet:
        opts.verbose = 0

    # debugging is fun
    if opts.debug:
        print(opts)
        return

    # run the workflow
    expression = get_expression_data(atlas=opts.atlas,
                                     atlas_info=opts.atlas_info,
                                     exact=opts.exact,
                                     tolerance=opts.tolerance,
                                     region_agg=opts.region_agg,
                                     agg_metric=opts.agg_metric,
                                     ibf_threshold=opts.ibf_threshold,
                                     probe_selection=opts.probe_selection,
                                     lr_mirror=opts.lr_mirror,
                                     gene_norm=opts.gene_norm,
                                     sample_norm=opts.sample_norm,
                                     corrected_mni=opts.corrected_mni,
                                     reannotated=opts.reannotated,
                                     return_counts=opts.save_counts,
                                     return_donors=opts.save_donors,
                                     donors=opts.donors,
                                     data_dir=opts.data_dir,
                                     verbose=opts.verbose)

    output_path = os.path.dirname(opts.output_file)
    fname_pref = os.path.splitext(os.path.basename(opts.output_file))[0]

    # WHY?!?
    if opts.stdout and not (opts.save_counts or opts.save_donors):
        expression.to_csv(sys.stdout)
        return

    # expand the tuple, if needed
    if opts.save_counts:
        expression, counts = expression
        counts_fname = os.path.join(output_path, fname_pref + '_counts.csv')
        lgr.info('Saving samples counts to {}'.format(counts_fname))
        counts.to_csv(counts_fname)

    # determine how best to save expression output files
    if opts.save_donors:
        if opts.donors == 'all':
            donors = list(donors.value_set('subj'))
        else:
            donors = [donors[f] for f in opts.donors]

        # save each donor dataframe as a separate file
        for donor, exp in zip(donors, expression):
            exp_fname = os.path.join(output_path,
                                     fname_pref + '_{}.csv'.format(donor))
            lgr.info('Saving donor {} info to {}'.format(donor, exp_fname))
            exp.to_csv(exp_fname)
    else:
        expression.to_csv(opts.output_file)


if __name__ == '__main__':
    raise RuntimeError('abagen/cli/run.py should not be run directly.\nPlease '
                       '`pip install` abagen and use the `abagen` command.')



# -*- coding: utf-8 -*-
"""
Functions to fetch mouse unionization (i.e., expression) data
"""

import itertools

from nibabel.volumeutils import Recoder
import numpy as np
import pandas as pd

from .io import fetch_allenref_structures, fetch_rubinov2015_structures
from .utils import _coerce_inputs, _make_api_query

# available attributes of unionization query
_UNIONIZATION_ATTRIBUTES = [
    'expression_density',
    'expression_energy',
    'sum_expressing_pixel_intensity',
    'sum_expressing_pixels',
    'sum_pixel_intensity',
    'sum_pixels',
    'voxel_energy_cv',
    'voxel_energy_mean'
]


def _get_experiments_from_gene(id=None, acronym=None, name=None,
                               slicing_direction='sagittal', verbose=False):
    """
    Fetches experiment IDs associated with specified gene(s)

    One of `id`, `acronym`, or `name` must be provided.

    Parameters
    ----------
    id : int, optional
        Numerical gene ID
    acronym : str, optional
        Short-form gene acronym (case sensitive)
    name : str, optional
        Full gene name (case sensitive)
    slicing_direction : {'sagittal', 'coronal'}, optional
        Slicing direction of brain tissue
    verbose : bool, optional
        Whether to print status messages. Default: False

    Returns
    -------
    experiments : list of int
        List of experiment IDs that probed the specified gene(s)
    """

    directions = ['sagittal', 'coronal']
    if slicing_direction not in directions:
        raise ValueError('Slicing_direction {} is invalid. Must be in {}.'
                         .format(slicing_direction, directions))

    criteria = [
        '[failed$eqfalse]',
        'products[id$eq1]',
        'genes{}'.format(_coerce_inputs(id=id, acronym=acronym, name=name)),
        'plane_of_section[name$eq{}]'.format(slicing_direction)
    ]

    info = _make_api_query('SectionDataSet', criteria=criteria,
                           attributes='data_sets.id', verbose=verbose)

    return [exp['id'] for exp in info]


def _get_unionization_from_experiment(experiment_id, structures=None,
                                      attributes=None, average=True,
                                      verbose=False):
    """
    Gets unionization data for provided experiment(s) `experiment_id`

    Parameters
    ----------
    experiment_id : int or list
        Numerical experiment ID. If multiple experiments are provided the
        requested `attributes` will be averaged across experiments
    structures : list, optional
        List of structures (id, acronym, or name) for which to get unionization
        information associated with provided `experiment_id`. If not specified
        uses structures documented in [1]_. Specifying either the id or name is
        recommended as acronyms are not unique to structures. Default: None
    attributes : str or list, optional
        Which attributes / information to obtain for the provided structure.
        See :func:`abagen.mouse.available_unionization_info` for list of
        available attributes to request. If not specified all available
        attributes will be returned. Default: None
    average : bool, optional
        Whether to average across experiments if `experiment_id` is provided as
        a list. Only experiments probing the same gene will be considered for
        averaging. Default: True
    verbose : bool, optional
        Whether to print status messages. Default: False

    Returns
    -------
    unionization : pandas.DataFrame
        Where columns are unionization attributes and the index corresponds to
        gene ids and strucuture ids

    References
    ----------
    .. [1] Rubinov, M., Ypma, R. J., Watson, C., & Bullmore, E. T. (2015).
       Wiring cost and topological participation of the mouse brain connectome.
       Proceedings of the National Academy of Sciences, 112(32), 10032-10037.
    """

    if isinstance(experiment_id, (str, int)):
        experiment_id = [experiment_id]

    if structures is None:
        # read default structure list (from Rubinov et al., 2015, PNAS)
        structures = fetch_rubinov2015_structures(entry_type='id')
    elif isinstance(structures, (str, int)):
        structures = [structures]

    # we need to coerce all provided structures to be integer ids, NOT strings
    # so fetch all available structures then recode them to ids
    if any(isinstance(f, str) for f in structures):
        structs = np.asarray(fetch_allenref_structures(verbose=False))
        structs = Recoder(structs.tolist(), fields=['acronym', 'id', 'name'])
        structures = list(set(structs.id.get(f) for f in structures))

    # determine which attributes to request; if we don't have to request all
    # of them then we can speed up the API call
    if attributes is None:
        attributes = ['expression_density']
    elif attributes == 'all':
        attributes = _UNIONIZATION_ATTRIBUTES
    elif isinstance(attributes, str):
        attributes = [attributes]

    includes = [
        'structure_unionizes', 'genes'
    ]
    criteria = [
        '[id$in{}]'.format(
            ','.join([str(f) for f in experiment_id])
        ),
        'products[id$eq1]',
        'structure_unionizes[structure_id$in{}]'.format(
            ','.join([str(f) for f in structures])
        ),
    ]
    req_attributes = [
        'id', 'structure_unionizes', 'structure_unionizes.structure_id'
    ] + [
        'structure_unionizes.' + attr for attr in attributes
    ]

    info = _make_api_query('SectionDataSet', includes=includes,
                           criteria=criteria, attributes=req_attributes,
                           verbose=verbose)

    for n, exp in enumerate(info):
        keep = exp['structure_unionizes']
        for struc in keep:
            struc['gene_id'] = exp['genes'][0]['id']
            struc['experiment_id'] = exp['id']
        info[n] = keep

    # construct data frame from requested unionization info
    info = pd.DataFrame(list(itertools.chain.from_iterable(info)))
    if average:
        info = info.groupby(['gene_id', 'structure_id']).mean()
    else:
        info = info.set_index(['gene_id', 'experiment_id', 'structure_id'])

    return info[attributes]


def available_unionization_info():
    """ Lists available attributes for `get_unionization_from_` functions
    """

    return _UNIONIZATION_ATTRIBUTES


def get_unionization_from_gene(id=None, acronym=None, name=None,
                               slicing_direction='sagittal', structures=None,
                               attributes=None, average=True, verbose=False):
    """
    Gets unionization data for provided gene(s)

    One of `id`, `acronym`, or `name` must be provided.

    Parameters
    ----------
    id : int, optional
        Numerical gene ID
    acronym : str, optional
        Short-form gene acronym (case sensitive)
    name : str, optional
        Full gene name (case sensitive)
    slicing_direction : {'sagittal', 'coronal'}, optional
        Slicing direction of brain tissue
    structures : list, optional
        List of structures (id, acronym, or name) for which to get unionization
        information associated with provided `experiment_id`. If not specified
        uses structures documented in [1]_. Specifying either the id or name is
        recommended as acronyms are not unique to structures. Default: None
    attributes : str or list, optional
        Which attributes / information to obtain for the provided gene. See
        :func:`abagen.mouse.available_gene_info` for list of available
        attributes to request. If not specified then only 'expression_density'
        will be returned. Specifying 'all' will return all information.
        Default: None
    average : bool, optional
        Whether to average across experiments if there are multiple experiments
        corresponding to any provided gene(s). Only experiments probing the
        same gene will be considered for averaging, and distinct structures
        will be retained. Default: True
    verbose : bool, optional
        Whether to print status messages. Default: False

    Returns
    -------
    unionization : pandas.DataFrame
        Where columns are unionization attributes and the index corresponds to
        strucuture and gene ids (if `experiments` is provided as a list
        with multiple genes). If `average=False`, `experiments` will also be a
        level in index

    Examples
    --------
    >>> from abagen import mouse
    >>> mouse.get_unionization_from_gene(acronym='Pdyn', structures=[22, 31])
                          expression_density
    gene_id structure_id
    18376   22                      0.024840
            31                      0.017199

    >>> mouse.get_unionization_from_gene(acronym=['Ace', 'Cd99'], structures=[22, 31])
                          expression_density
    gene_id structure_id
    11210   22                      0.001283
            31                      0.001427
    163028  22                      0.067537
            31                      0.056442

    References
    ----------
    .. [1] Rubinov, M., Ypma, R. J., Watson, C., & Bullmore, E. T. (2015).
       Wiring cost and topological participation of the mouse brain connectome.
       Proceedings of the National Academy of Sciences, 112(32), 10032-10037.
    """  # noqa

    directions = ['sagittal', 'coronal']
    if slicing_direction not in directions:
        raise ValueError('Slicing_direction {} is invalid. Must be in {}.'
                         .format(slicing_direction, directions))

    if structures is None:
        # read default structure list (from Rubinov et al., 2015, PNAS)
        structures = fetch_rubinov2015_structures(entry_type='id')
    elif isinstance(structures, (str, int)):
        structures = [structures]

    # we need to coerce all provided structures to be integer ids, NOT strings
    # so fetch all available structures then recode them to ids
    if any(isinstance(f, str) for f in structures):
        structs = np.asarray(fetch_allenref_structures(verbose=False))
        structs = Recoder(structs.tolist(), fields=['acronym', 'id', 'name'])
        structures = list(set(structs.id.get(f) for f in structures))

    exp_ids = _get_experiments_from_gene(id=id, acronym=acronym, name=name,
                                         slicing_direction=slicing_direction,
                                         verbose=verbose)

    data = _get_unionization_from_experiment(exp_ids, structures=structures,
                                             attributes=attributes,
                                             average=average, verbose=verbose)

    return data



# -*- coding: utf-8 -*-
"""
Functions for downloading the Allen Brain Atlas human microarray dataset.
"""

import multiprocessing as mp
import os
from pkg_resources import resource_filename

from nibabel.volumeutils import Recoder
import pandas as pd

from .. import io
from .utils import _get_dataset_dir, _fetch_files

WELL_KNOWN_IDS = Recoder(
    (('9861', 'H0351.2001', '178238387', '157722636', '157722638'),
     ('10021', 'H0351.2002', '178238373', '157723301', '157723303'),
     ('12876', 'H0351.1009', '178238359', '157722290', '157722292'),
     ('15496', 'H0351.1015', '178238266', '162021642', '162021644'),
     ('14380', 'H0351.1012', '178238316', '157721937', '157721939'),
     ('15697', 'H0351.1016', '178236545', '157682966', '157682968')),
    fields=('subj', 'uid', 'url', 't1w', 't2w')
)
VALID_DONORS = sorted(WELL_KNOWN_IDS.value_set('subj')
                      | WELL_KNOWN_IDS.value_set('uid'))


def check_donors(donors, default='12876', valid=VALID_DONORS):
    """
    Checks that provided `donors` are valid

    Parameters
    ----------
    donors : list of str
        List of donors to download; can be either donor number or UID. Can also
        specify 'all' to download all available donors. If 'None' is provided
        then `default` will be used.
    default : str, optional
        Default donor to use if `donors` is None. Default: '12876'
    valid : list of str, optional
        List of valid donnor numbers and UIDs. Default: :obj:`VALID_DONORS`

    Returns
    -------
    donors : list of str
        Donor subject IDs
    """

    if donors is None:
        donors = [default]
    elif donors == 'all':
        donors = valid
    elif isinstance(donors, str):
        donors = [donors]

    donors = list(donors).copy()
    for n, sub_id in enumerate(donors):
        if sub_id not in valid:
            raise ValueError('Invalid subject id: {0}. Subjects must in: {1}.'
                             .format(sub_id, valid))
        donors[n] = WELL_KNOWN_IDS[sub_id]  # convert to ID system
    donors = sorted(set(donors), key=lambda x: int(x))

    return donors


def fetch_microarray(data_dir=None, donors=None, resume=True, verbose=1,
                     convert=True, n_proc=1):
    """
    Downloads the Allen Human Brain Atlas microarray expression dataset

    Parameters
    ----------
    data_dir : str, optional
        Directory where data should be downloaded and unpacked. Default: $HOME/
        abagen-data
    donors : list, optional
        List of donors to download; can be either donor number or UID. Can also
        specify 'all' to download all available donors. Default: 12876
    resume : bool, optional
        Whether to resume download of a partly-downloaded file. Default: True
    verbose : int, optional
        Verbosity level (0 means no message). Default: 1
    convert : bool, optional
        Whether to convert downloaded CSV files into parquet format for faster
        loading in the future; only available if ``fastparquet`` and ``python-
        snappy`` are installed. Default: True
    n_proc : int, optional
        Number of processes to parallelize download if multiple donors are
        specified. Default: 1

    Returns
    -------
    data : dict
        Dictionary with keys ['microarray', 'ontology', 'pacall', 'probes',
        'annotation'], where corresponding values are lists of filepaths to
        downloaded CSV files.

    References
    ----------
    Hawrylycz, M. J., Lein, E. S., Guillozet-Bongaarts, A. L., Shen, E. H., Ng,
    L., Miller, J. A., ... & Abajian, C. (2012). An anatomically comprehensive
    atlas of the adult human brain transcriptome. Nature, 489(7416), 391.
    """

    url = "https://human.brain-map.org/api/v2/well_known_file_download/{}"

    dataset_name = 'microarray'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir,
                                verbose=verbose)

    sub_files = ('MicroarrayExpression.csv',
                 'Ontology.csv', 'PACall.csv',
                 'Probes.csv', 'SampleAnnot.csv')
    n_files = len(sub_files)
    donors = check_donors(donors)

    if n_proc < 0:
        n_proc = mp.cpu_count() + n_proc + 1

    files = [
        [(os.path.join('normalized_microarray_donor{}'.format(sub), fname),
            url.format(WELL_KNOWN_IDS.url[sub]),
            dict(uncompress=True,
                 move=os.path.join('normalized_microarray_donor{}'.format(sub),
                                   'donor{}.zip'.format(sub))))
         for fname in sub_files]
        for sub in donors
    ]

    if n_proc > 1:
        with mp.Pool(n_proc) as pool:
            results = [pool.apply_async(_fetch_files,
                                        (data_dir, f),
                                        dict(resume=resume, verbose=verbose))
                       for f in files]
            # flatten outputs into single list
            files = [fn for res in results for fn in res.get()]
    else:
        # flatten list of lists into single list
        files = [fn for f in files for fn in f]
        files = _fetch_files(data_dir, files, resume=resume, verbose=verbose)

    # if we want to convert files to parquet format it's good to do that now
    # this step is _already_ super long, so an extra 1-2 minutes is negligible
    if convert and io.use_parq:
        for fn in files[0::n_files] + files[2::n_files]:
            io._make_parquet(fn, convert_only=True)

    keys = ['microarray', 'ontology', 'pacall', 'probes', 'annotation']
    return {
        donor: dict(zip(keys, files[k:k + n_files]))
        for k, donor in zip(range(0, len(files), n_files), donors)
    }


def fetch_rnaseq(data_dir=None, donors=None, resume=True, verbose=1):
    """
    Downloads RNA-sequencing data from the Allen Human Brain Atlas

    Parameters
    ----------
    data_dir : str, optional
        Directory where data should be downloaded and unpacked. Default:
        current directory
    donors : list, optional
        List of donors to download; can be either donor number or UID. Can also
        specify 'all' to download all available donors (two). Default: 9861
    resume : bool, optional
        Whether to resume download of a partly-downloaded file. Default: True
    verbose : int, optional
        Verbosity level (0 means no message). Default: 1

    Returns
    -------
    data : dict
        Dictionary with keys ['counts', 'tpm', 'ontology', 'genes',
        'annotation'], where corresponding values are lists of filepaths to
        downloaded CSV files.

    References
    ----------
    Hawrylycz, M. J., Lein, E. S., Guillozet-Bongaarts, A. L., Shen, E. H., Ng,
    L., Miller, J. A., ... & Abajian, C. (2012). An anatomically comprehensive
    atlas of the adult human brain transcriptome. Nature, 489(7416), 391.
    """

    url = "https://human.brain-map.org/api/v2/well_known_file_download/{}"
    well_known_ids = {
        '9861': '278447594',
        '10021': '278448166'
    }

    dataset_name = 'rnaseq'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir,
                                verbose=verbose)

    sub_files = ('Genes.csv', 'Ontology.csv',
                 'RNAseqCounts.csv', 'RNAseqTPM.csv', 'SampleAnnot.csv')
    n_files = len(sub_files)
    valid = ['9861', '10021', 'H0351.2001', 'H0351.2002']
    donors = sorted(set(check_donors(donors, default=valid[0])) & set(valid),
                    key=lambda x: int(x))

    files = [
        [(os.path.join('rnaseq_donor{}'.format(sub), fname),
            url.format(well_known_ids[sub]),
            dict(uncompress=True,
                 move=os.path.join('rnaseq_donor{}'.format(sub),
                                   'donor{}.zip'.format(sub))))
         for fname in sub_files]
        for sub in donors
    ]

    files = [fn for f in files for fn in f]
    files = _fetch_files(data_dir, files, resume=resume, verbose=verbose)

    keys = ['genes', 'ontology', 'counts', 'tpm', 'annotation']
    return {
        donor: dict(zip(keys, files[k:k + n_files]))
        for k, donor in zip(range(0, len(files), n_files), donors)
    }


def fetch_raw_mri(data_dir=None, donors=None, resume=True, verbose=1):
    """
    Downloads the "raw" Allen Human Brain Atlas T1w/T2w MRI images

    Parameters
    ----------
    data_dir : str, optional
        Directory where data should be downloaded and unpacked. Default: $HOME/
        abagen-data
    donors : list, optional
        List of donors to download; can be either donor number or UID. Can also
        specify 'all' to download all available donors. Default: 12876
    resume : bool, optional
        Whether to resume download of a partly-downloaded file. Default: True
    verbose : int, optional
        Verbosity level (0 means no message). Default: 1

    Returns
    -------
    mris : dict
        Dictionary with keys ['t1w', 't2w'], where corresponding values are
        lists of filepaths to downloaded Nifti files
    """

    url = "https://human.brain-map.org/api/v2/well_known_file_download/{}"

    dataset_name = 'mri'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir,
                                verbose=verbose)

    sub_files = dict(t1w='T1.nii.gz', t2w='T2.nii.gz')
    n_files = len(sub_files)
    donors = check_donors(donors)

    files = [
        (os.path.join('mri_donor{}'.format(sub), fname),
         url.format(getattr(WELL_KNOWN_IDS, img)[sub]),
         dict(move=os.path.join('mri_donor{}'.format(sub),
                                fname)))
        for sub in donors
        for img, fname in sub_files.items()
    ]

    files = _fetch_files(data_dir, files, resume=resume, verbose=verbose)

    return {
        donor: dict(zip(sub_files.keys(), files[k:k + n_files]))
        for k, donor in zip(range(0, len(files), n_files), donors)
    }


def fetch_freesurfer(data_dir=None, donors=None, resume=True, verbose=1):
    """
    Downloads FreeSurfer reconstructions of the Allen Human Brain Atlas MRIs

    Parameters
    ----------
    data_dir : str, optional
        Directory where data should be downloaded and unpacked. Default: $HOME/
        abagen-data
    donors : list, optional
        List of donors to download; can be either donor number or UID. Can also
        specify 'all' to download all available donors. Default: 12876
    resume : bool, optional
        Whether to resume download of a partly-downloaded file. Default: True
    verbose : int, optional
        Verbosity level (0 means no message). Default: 1

    Returns
    -------
    freesurfer : list
        List to FreeSurfer directories for requested `donors`

    References
    ----------
    Romero-Garcia, R., Whitaker, K., Vasa, F., Seidlitz, J., Shinn, M., Fonagy,
    P., Jones, P., et al. (2017). Data supporting NSPN publication "Structural
    covariance networks are coupled to expression of genes enriched in
    supragranular layers of the human cortex " [Dataset].
    https://doi.org/10.17863/CAM.11392
    """

    url = "https://www.repository.cam.ac.uk/bitstream/handle/1810/265272/" \
          "donor{}.zip"
    dataset_name = 'freesurfer'
    data_dir = _get_dataset_dir(dataset_name, data_dir=data_dir,
                                verbose=verbose)

    donors = check_donors(donors)
    files = [
        ('donor{}'.format(sub),
         url.format(sub),
         dict(uncompress=True,
              move=os.path.join('freesurfer.tar.gz')))
        for sub in donors
    ]

    files = _fetch_files(data_dir, files, resume=resume, verbose=verbose)

    return {
        donor: files[k]
        for k, donor in enumerate(donors)
    }


def fetch_desikan_killiany(native=False, *args, **kwargs):
    """
    Fetches Desikan-Killiany atlas shipped with `abagen`

    Parameters
    ----------
    native : bool, optional
        Whether to return group-level atlas in MNI space (False) instead of
        individualized atlases in donor native space (True). Default: False

    Returns
    -------
    atlas : dict
        Dictionary with keys ['image', 'info'] pointing to atlas image
        (.nii.gz) and information (.csv) files. If ``native=True`` then 'image'
        is a dictionary where keys are donor IDs and values are image paths.

    References
    ----------
    Desikan, R. S., Ségonne, F., Fischl, B., Quinn, B. T., Dickerson, B. C.,
    Blacker, D., ... & Albert, M. S. (2006). An automated labeling system
    for subdividing the human cerebral cortex on MRI scans into gyral based
    regions of interest. Neuroimage, 31(3), 968-980.

    Examples
    --------
    >>> import abagen
    >>> atlas = abagen.fetch_desikan_killiany()
    >>> print(atlas['image'])  # doctest: +ELLIPSIS
    /.../abagen/data/atlas-desikankilliany.nii.gz
    >>> print(atlas['info'])  # doctest: +ELLIPSIS
    /.../abagen/data/atlas-desikankilliany.csv

    When fetching native-space atlases, `atlas['image']` will be a dictionary
    where the keys are donor IDs and the values are paths to the donor-specific
    atlases:

    >>> atlas = abagen.fetch_desikan_killiany(native=True)
    >>> print(atlas['image'].keys())
    dict_keys(['9861', '10021', '12876', '14380', '15496', '15697'])
    >>> print(atlas['image']['9861'])  # doctest: +ELLIPSIS
    /.../abagen/data/native_dk/9861/atlas-desikankilliany.nii.gz
    """

    # grab resource filenames
    if not native:
        img = resource_filename('abagen', 'data/atlas-desikankilliany.nii.gz')
    else:
        img = dict()
        for donor in check_donors('all'):
            imgpath = f'data/native_dk/{donor}/atlas-desikankilliany.nii.gz'
            img[donor] = resource_filename('abagen', imgpath)
    info = resource_filename('abagen', 'data/atlas-desikankilliany.csv')

    return dict(image=img, info=info)


def fetch_gene_group(group):
    """
    Return list of gene acronyms belonging to provided `group`

    Groups are defined as in [DS1]_

    Parameters
    ----------
    group : {'brain', 'neuron', 'oligodendrocyte', 'synaptome', 'layers'}
        Desired gene group

    Returns
    -------
    genes : list of str
        List of gene acronyms

    References
    ----------
    .. [DS1] Burt, J. B., Demirtaş, M., Eckner, W. J., Navejar, N. M., Ji, J.
       L., Martin, W. J., ... & Murray, J. D. (2018). Hierarchy of
       transcriptomic specialization across human cortex captured by
       structural neuroimaging topography. Nature neuroscience, 21(9), 1251.
    """

    groups = ['brain', 'neuron', 'oligodendrocyte', 'synaptome', 'layers']
    if group.lower() not in groups:
        raise ValueError('Provided group {} not one of the available gene '
                         'groups: {}'.format(group, groups))

    group = group.lower()
    fn = resource_filename('abagen', 'data/burt2018_natneuro.csv')
    genes = pd.read_csv(fn).query('group == "{}"'.format(group))['acronym']

    return sorted(list(genes))


def fetch_donor_info():
    """
    Returns dataframe with donor demographic information

    Returns
    -------
    info : pandas.DataFrame
        With columns ['donor', 'age', 'sex', 'ethnicity', 'medical_conditions',
        'post_mortem_interval_hours'] detailing basic demographic info about
        donors
    """

    fn = resource_filename('abagen', 'data/donor_info.csv')
    donors = pd.read_csv(fn)

    return donors



# -*- coding: utf-8 -*-

import random
import pytest
from abagen.mouse import io, gene

# number of tests to make
RANDOM = random.Random(1234)
GENES = io.fetch_allenref_genes()


def test_get_gene_info():
    samples = GENES.loc[RANDOM.sample(range(len(GENES)), 10)]

    acronyms = gene.get_gene_info(id=samples['id'], attributes='acronym')
    assert sorted(acronyms['acronym']) == sorted(samples['acronym'])

    names = gene.get_gene_info(acronym=samples['acronym'], attributes='name')
    assert sorted(names['name']) == sorted(samples['name'])

    with pytest.raises(ValueError):
        gene.get_gene_info(acronym=samples['acronym'], attributes='invalid')

    # multiple attributes
    info = gene.get_gene_info(id=samples['id'], attributes=['name', 'acronym'])
    assert sorted(acronyms['acronym']) == sorted(samples['acronym'])
    assert sorted(info['name']) == sorted(samples['name'])

    # all attributes
    info = gene.get_gene_info(acronym=samples['acronym'])
    assert len(info.columns) == 15

    # exception: invalid gene identifiers
    with pytest.raises(ValueError):
        gene.get_gene_info(id=-100000)
    with pytest.raises(ValueError):
        gene.get_gene_info(acronym='notagene')



# -*- coding: utf-8 -*-
"""
Functions to make mouse structure queries and manipulations
"""

import itertools
import re
import pandas as pd
from .utils import _coerce_inputs, _make_api_query

# available attributes of structure query
_STRUCTURE_ATTRIBUTES = [
    'acronym',
    'atlas_id',
    'color_hex_triplet',
    'depth',
    'graph_id',
    'graph_order',
    'hemisphere_id',
    'id',
    'name',
    'neuro_name_structure_id',
    'neuro_name_structure_id_path',
    'ontology_id',
    'parent_structure_id',
    'safe_name',
    'sphinx_id',
    'structure_id_path',
    'weight'
]


def available_structure_info():
    """ Lists available attributes for :func:`abagen.mouse.get_structure_info`
    """

    return _STRUCTURE_ATTRIBUTES.copy()


def get_structure_info(id=None, acronym=None, name=None, attributes=None,
                       verbose=False):
    """
    Queries Allen API for information about given gene

    One of `structure_id`, `structure_acronym`, or `structure_name` must be
    provided.

    Parameters
    ----------
    id : int, optional
        Numerical structure ID
    acronym : str, optional
        Short-form structure acronym (case sensitive)
    name : str, optional
        Full structure name (case sensitive)
    attributes : str or list, optional
        Which attributes / information to obtain for the provided structure.
        See :func:`abagen.mouse.available_structure_info` for list of available
        attributes to request. If not specified all available attributes will
        be returned. Default: None
    verbose : bool, optional
        Whether to print status messages. Default: False

    Returns
    -------
    info : pandas.DataFrame
        Where columns are the requested attributes and index is the provided
        structural identifier type (e.g., 'id', 'acronym', 'name')

    Raises
    ------
    ValueError
        The provided structure is invalid

    Examples
    --------
    Get the full names of structures 22 and 1018:
    >>> from abagen import mouse
    >>> mouse.get_structure_info(id=[22, 1018], attributes=['acronym', 'name'])
         acronym                                  name
    id
    22      PTLp  Posterior parietal association areas
    1018    AUDv                 Ventral auditory area

    """

    criteria = [
        _coerce_inputs(id=id, acronym=acronym, name=name),
        'ontology[id$eq1]'
    ]
    provided = re.search(r'\[(\S+)\$', criteria[0]).group(1)

    # determine which attributes to request; if we don't have to request all
    # of them then we can speed up the API call
    if attributes is None:
        attributes = _STRUCTURE_ATTRIBUTES
    elif isinstance(attributes, str):
        attributes = [attributes]
    attributes = [a for a in attributes if a not in provided]

    for attr in attributes:
        if attr not in _STRUCTURE_ATTRIBUTES:
            raise ValueError('Provided attribute "{}" is invalid; please '
                             'check valid attributes with '
                             'abagen.mouse.available_structure_info().'
                             .format(attr))

    info = _make_api_query('Structure', criteria=criteria,
                           attributes=attributes + [provided], verbose=verbose)
    info = pd.DataFrame(info).set_index(provided)[attributes]

    return info.sort_index()


def get_structure_coordinates(id=None, acronym=None, name=None,
                              reference_space='sagittal', verbose=False):
    """
    Finds xyz coordinates of provided structure(s) in `reference_space`

    Parameters
    ----------
    id : int, optional
        Numerical structure ID
    acronym : str, optional
        Short-form structure acronym (case sensitive)
    name : str, optional
        Full structure name (case sensitive)
    reference_space : {'sagittal', 'coronal'}, optional
        Reference space from which to extract coordinates. Default: 'sagittal'
    verbose : bool, optional
        Whether to print status messages. Default: False

    Returns
    -------
    coords : pandas.DataFrame
        With columns ['structure_id', 'x', 'y', 'z']

    Examples
    --------
    Get the coordinates of structure 1018:
    >>> from abagen import mouse
    >>> mouse.get_structure_coordinates(id=1018)
       structure_id     x     y     z
    0          1018  7800  3400  1050
    """

    spaces = {'sagittal': 10, 'coronal': 9}
    if reference_space not in spaces:
        raise ValueError('Reference_space {} is invalid. Must be in {}.'
                         .format(reference_space, list(spaces.keys())))
    space = spaces[reference_space]

    includes = 'structure_centers'
    criteria = [
        _coerce_inputs(id=id, acronym=acronym, name=name),
        'ontology[id$eq1]',
        'structure_centers[reference_space_id$eq{}]'.format(space)
    ]
    attributes = [
        'structure_centers', 'structure_centers.structure_id',
        'structure_centers.x', 'structure_centers.y', 'structure_centers.z'
    ]

    info = _make_api_query('Structure', includes=includes, criteria=criteria,
                           attributes=attributes, verbose=verbose)
    info = [struct['structure_centers'] for struct in info]
    coords = pd.DataFrame(list(itertools.chain.from_iterable(info)))

    return coords



# -*- coding: utf-8 -*-
"""
Functions for loading the various files associated with the AHBA microarray and
RNAseq datasets.

This also contains functionality for optionally converting the downloaded CSV
files to parquet format, which provides much faster I/O access / quicker load
times.
"""

import os.path as op
import pandas as pd
try:
    eng = pd.io.parquet.get_engine('fastparquet')
    assert 'SNAPPY' in eng.api.compression.compressions
    use_parq = True
# pandas version too low OR don't have fastparquet installed
except (AttributeError, ImportError, AssertionError):
    use_parq = False


def _make_parquet(fname, convert_only=False):
    """
    Loads `fname`, converting to parquet file if it does not already exist

    Parameters
    ----------
    fname : str
        Path to data file
    convert_only : bool, optional
        Check if parquet version of `fname` exists and convert if it doesn't;
        if it does, just return. Default: False

    Returns
    -------
    data : pandas.DataFrame
        Data loaded from `fname`
    """

    # get ideal parquet filename
    parqname = fname.rpartition('.csv')[0] + '.parq'

    # if it exists, load it as parquet
    if op.exists(parqname):
        if convert_only:
            return
        data = pd.read_parquet(parqname, engine='fastparquet')
    # otherwise, load CSV and save to parquet
    else:
        data = pd.read_csv(fname, header=None)
        data.columns = data.columns.astype(str)
        data.to_parquet(parqname, engine='fastparquet')
        if convert_only:
            return

    return data


def read_microarray(fname, copy=False, parquet=True):
    """
    Loads MicroarrayExpression.csv file found at `fname`

    Microarray files contain raw expression data for all the tissue samples
    taken from a single donor across all genetic probes.

    Parameters
    ----------
    fname : str
        Path to MicroarrayExpression.csv file
    copy : bool, optional
        Whether to return a copy if `fname` is a pre-loaded pandas.Dataframe.
        Default: False
    parquet : bool, optional
        Whether to load data from parquet file instead of CSV. If a parquet
        file does not already exist then one will be created for faster loading
        in the future. Only available if ``fastparquet`` and ``python-snappy``
        module are installed. Default: True

    Returns
    -------
    microarray : (P, S) pandas.DataFrame
        Dataframe containing microarray expression data, where `P` is probes
        and `S` is samples. The row index is the unique probe ID assigned
        during processing, which can be used to match data to the information
        obtained with :func:`read_probes`. The column index is the unique
        sample ID (integer, beginning at 0) which can be used to match data to
        the information obtained with :func:`read_annotation`.
    """

    try:
        if use_parq and parquet:
            data = _make_parquet(fname, convert_only=False)
            data = data.set_index('0')
        else:
            data = pd.read_csv(fname, header=None, index_col=0)
        data.index.name = 'probe_id'
        data.columns = pd.Series(range(len(data.columns)), name='sample_id')
    except (AttributeError, ValueError, TypeError):
        if not isinstance(fname, pd.DataFrame):
            raise TypeError('Provided fname must be filepath to Microarray'
                            'Expression.csv file from Allen Human Brain '
                            'Atlas.')
        data = fname.copy() if copy else fname

    return data


def read_ontology(fname, copy=False):
    """
    Loads Ontology.csv file found at `fname`

    Ontology files contain information on the anatomical delineations used by
    the Allen Institute when obtaining samples from donor brains, and are used
    in their online Brain Viewer to colorize regions. These files should be the
    same for every donors.

    This information can be used to ensure that microarray samples are
    appropriately matched to anatomical regions.

    Parameters
    ----------
    fname : str
        Path to Ontology.csv file
    copy : bool, optional
        Whether to return a copy if `fname` is a pre-loaded pandas.Dataframe.
        Default: False

    Returns
    -------
    ontology : (R, 8) pandas.DataFrame
        Dataframe containing ontology information for `R` anatomical regions
        used by the Allen Institute. Columns include: 'id', 'acronym', 'name',
        'parent_structure_id', 'hemisphere', 'graph_order',
        'structure_id_path', and 'color_hex_triplet'.
    """

    try:
        data = pd.read_csv(fname)
    except (ValueError, TypeError):
        if not isinstance(fname, pd.DataFrame):
            raise TypeError('Provided fname must be filepath to Ontology.csv '
                            'file from Allen Human Brain Atlas.')
        data = fname.copy() if copy else fname

    return data


def read_pacall(fname, copy=False, parquet=True):
    """
    Loads PACall.csv file found at `fname`

    PA files contain a present/absent flag indicating whether the corresponding
    probe's expression is above background noise. It is set to 1 when both of
    the following conditions are met:

        1. The mean signal of the probe's expression is significantly different
           from the corresponding background, as assessed by a 2-sided t-test
           where p < 0.01, and
        2. The difference between the background subtracted signal and the
           background is significant (> 2.6 * background standard deviation).

    This information can be used to discard "noisy" probes that might not be
    contributing high-quality expression information.

    Parameters
    ----------
    fname : str
        Path to PACall.csv file
    copy : bool, optional
        Whether to return a copy if `fname` is a pre-loaded pandas.Dataframe.
        Default: False
    parquet : bool, optional
        Whether to load data from parquet file instead of CSV. If a parquet
        file does not already exist then one will be created for faster loading
        in the future. Only available if ``fastparquet`` and ``python-snappy``
        module are installed. Default: True

    Returns
    -------
    pacall : (P, S) pandas.DataFrame
        Dataframe containing a binary indicator determining whether expression
        information for each probe exceeded background noise in a given sample,
        where `P` is probes and `S` is samples. The row index is the unique
        probe ID assigned during processing, which can be used to match data to
        the information obtained with :func:`read_probes`. The column index is
        the unique sample ID (integer, beginning at 1) which can be used to
        match data to the information obtained with :func:`read_annotation`.
    """

    try:
        if use_parq and parquet:
            data = _make_parquet(fname, convert_only=False)
            data = data.set_index('0')
        else:
            data = pd.read_csv(fname, header=None, index_col=0)
        data.index.name = 'probe_id'
        data.columns = pd.Series(range(len(data.columns)), name='sample_id')
    except (AttributeError, ValueError, TypeError):
        if not isinstance(fname, pd.DataFrame):
            raise TypeError('Provided fname must be filepath to PACall.csv '
                            'file from Allen Human Brain Atlas.')
        data = fname.copy() if copy else fname

    return data


def read_probes(fname, copy=False):
    """
    Loads Probes.csv file found at `fname`

    Probe files contain metadata on all genetic probes used in the AHBA data.
    These files should be the same for every donor.

    This information can be used to e.g., query expression data for certain
    genes, collapse data across probes from the same gene, etc.

    Parameters
    ----------
    fname : str
        Path to Probes.csv file
    copy : bool, optional
        Whether to return a copy if `fname` is a pre-loaded pandas.Dataframe.
        Default: False

    Returns
    -------
    probes : (P, 6) pandas.DataFrame
        Dataframe containing information for `P` genetic probes. The row index
        is the unique probe ID assigned during processing, which can be used to
        match metadata to information obtained with :func:`read_microarray` and
        :func:`read_pacall`. Columns include 'probe_name', 'gene_id',
        'gene_symbol', 'gene_name', 'entrez_id', and 'chromosome'.
    """

    try:
        data = pd.read_csv(fname, index_col=0)
    except (ValueError, TypeError):
        if not isinstance(fname, pd.DataFrame):
            raise TypeError('Provided fname must be filepath to Probes.csv '
                            'file from Allen Human Brain Atlas.')
        data = fname.copy() if copy else fname

    return data


def read_annotation(fname, copy=False):
    """
    Loads SampleAnnot.csv file found at `fname`

    Sample annotation files contain metadata on all the tissue samples taken
    from a single donor brain, including the spatial location of the samples.

    This information can be used to combine samples within the same anatomical
    region across donors.

    Parameters
    ----------
    fname : str
        Path to SampleAnnot.csv file
    copy : bool, optional
        Whether to return a copy if `fname` is a pre-loaded pandas.Dataframe.
        Default: False

    Returns
    -------
    annotation : (S, 13) pandas.DataFrame
        Dataframe containing structural information on `S` samples. The row
        index is the unique sample ID (integer, beginning with 1) which can be
        used to match data to the information obtained with e.g.,
        :func:`read_microarray`.

    Notes
    -----
    If the provided annotation file is from microarray expression data
    (obtained by, e.g., `abagen.fetch_microarray()`), then the returned
    DataFrame will have the following columns: 'structure_id', 'slab_num',
    'well_id', 'slab_type', 'structure_acronym', 'structure_name',
    'polygon_id', 'mri_voxel_x', 'mri_voxel_y', 'mri_voxel_z', 'mni_x',
    'mni_y', 'mni_z'.

    If the provided annotation file is from RNAseq data (obtained by, e.g.,
    `abagen.fetch_rnaseq()`), then the returned DataFrame will have the
    following columns: 'RNAseq_sample_name', 'replicate_sample', 'sample_name',
    'well_id', 'microarray_run_id', 'ontology_color', 'main_structure',
    'sub_structure', 'structure_id', 'structure_acronym', 'hemisphere',
    'brain', 'million_clusters', 'clip_percentage', 'RIN_RNA_squality',
    'rnaseq_run_id', 'A.Pct', 'C.Pct', 'G.Pct', 'T.Pct', 'N.Pct'
    """

    mapper = dict(
        ontology_structure_id='structure_id',
        ontology_structure_acronym='structure_acronym'
    )

    try:
        data = pd.read_csv(fname)
        data.index.name = 'sample_id'
    except (ValueError, TypeError):
        if not isinstance(fname, pd.DataFrame):
            raise TypeError('Provided fname must be filepath to Annotation'
                            '.csv file from Allen Human Brain Atlas.')
        data = fname.copy() if copy else fname

    data.rename(mapper, axis='columns', inplace=True, errors='ignore')

    return data


def read_tpm(fname, copy=False):
    """
    Loads RNAseqTPM.csv file found at `fname`

    RNAseq TPM files contain TPM values for all the tissue samples taken from a
    single donor across all genes. TPM values are scaled fragment (read) counts
    derived using RSEM.

    Parameters
    ----------
    fname : str
        Path to RNAseqTPM.csv file
    copy : bool, optional
        Whether to return a copy if `fname` is a pre-loaded pandas.Dataframe.
        Default: False

    Returns
    -------
    tpm : (G, S) pandas.DataFrame
        Dataframe containing RNAseq TPM expression data, where `G` is genes
        and `S` is samples. The row index is the unique gene symbol assigned
        during processing, which can be used to match data to the information
        obtained with :func:`read_genes`. The column index is the unique
        sample ID (integer, beginning at 0) which can be used to match data to
        the information obtained with :func:`read_annotation`.
    """

    try:
        data = pd.read_csv(fname, header=None, index_col=0)
        data.index.name = 'gene_symbol'
        data.columns = pd.Series(range(len(data.columns)), name='sample_id')
    except (ValueError, TypeError):
        if not isinstance(fname, pd.DataFrame):
            raise TypeError('Provided fname must be filepath to RNAseqTPM'
                            '.csv file from Allen Human Brain Atlas.')
        data = fname.copy() if copy else fname

    return data


def read_counts(fname, copy=False):
    """
    Loads RNAseqCounts.csv file found at `fname`

    RNAseq count files contain fragment counts for all the tissue samples taken
    from a single donor across all genes. Fragment counts can be fractional, as
    ambiguous reads are distributed between relevant transcripts. For present /
    absent calling, a value of zero indicates no transcript was seen.

    Parameters
    ----------
    fname : str
        Path to RNAseqCounts.csv file
    copy : bool, optional
        Whether to return a copy if `fname` is a pre-loaded pandas.Dataframe.
        Default: False

    Returns
    -------
    tpm : (G, S) pandas.DataFrame
        Dataframe containing RNAseq count expression data, where `G` is genes
        and `S` is samples. The row index is the unique gene symbol assigned
        during processing, which can be used to match data to the information
        obtained with :func:`read_genes`. The column index is the unique
        sample ID (integer, beginning at 0) which can be used to match data to
        the information obtained with :func:`read_annotation`.
    """

    try:
        data = pd.read_csv(fname, header=None, index_col=0)
        data.index.name = 'gene_symbol'
        data.columns = pd.Series(range(len(data.columns)), name='sample_id')
    except (ValueError, TypeError):
        if not isinstance(fname, pd.DataFrame):
            raise TypeError('Provided fname must be filepath to RNAseqCounts'
                            '.csv file from Allen Human Brain Atlas.')
        data = fname.copy() if copy else fname

    return data


def read_genes(fname, copy=False):
    """
    Loads Genes.csv file found at `fname`

    Genes files contain metadata on all genes used in the RNAseq AHBA data.
    These files should be the same for every donor.

    Parameters
    ----------
    fname : str
        Path to Genes.csv file
    copy : bool, optional
        Whether to return a copy if `fname` is a pre-loaded pandas.Dataframe.
        Default: False

    Returns
    -------
    genes : (G, 11) pandas.DataFrame
        Dataframe containing information for `G` unique genes. The row index
        is the unique gene symbol which can be used to match metadata to
        information obtained with :func:`read_tpm` and :func:`read_counts`.
        Columns include 'gene_id', 'entrez_id', 'chromosome', 'strand',
        'number_of_transcripts', 'median_transcriptome_length',
        'median_genome_length', 'median_number_of_exons', 'median_gene_start',
        and 'median_gene_end'
    """

    try:
        data = pd.read_csv(fname, index_col=0)
    except (ValueError, TypeError):
        if not isinstance(fname, pd.DataFrame):
            raise TypeError('Provided fname must be filepath to Annotation'
                            '.csv file from Allen Human Brain Atlas.')
        data = fname.copy() if copy else fname

    return data



# -*- coding: utf-8 -*-
"""
Tests for abagen.utils module
"""

import numpy as np
import pandas as pd
import pytest

from abagen import utils


@pytest.mark.xfail
def test_check_dict():
    assert False


@pytest.mark.xfail
def test_flatten_dict():
    assert False


@pytest.mark.xfail
def test_first_entry():
    assert False


def test_leftify_atlas(atlas):
    out = utils.leftify_atlas(atlas['image'])
    assert len(np.unique(out.dataobj)) == 44
    assert np.all(np.asarray(out.dataobj)[98:] == 0)


def test_check_img(atlas):
    # some really basic, silly checks
    out = utils.check_img(atlas['image'])
    assert out.header.get_data_dtype() == np.dtype('int32')
    assert len(out.shape) == 3


def test_check_atlas_info(atlas):
    # general usage (providing two filenames) works as expected
    out = utils.check_atlas_info(atlas['image'], atlas['info'])
    assert all(out.columns == ['label', 'hemisphere', 'structure'])
    assert out.index.name == 'id'

    # can accept dataframe as input
    atlas_df = pd.read_csv(atlas['info'])
    out2 = utils.check_atlas_info(atlas['image'], atlas_df)
    pd.testing.assert_frame_equal(out, out2)

    # setting ID as index of dataframe is acceptable usage
    atlas_df = atlas_df.set_index('id')
    out3 = utils.check_atlas_info(atlas['image'], atlas_df)
    pd.testing.assert_frame_equal(out, out3)

    # check that coercion of different hemisphere designations works
    atlas_df.loc[atlas_df['hemisphere'] == "L", 'hemisphere'] = "lh"
    atlas_df.loc[atlas_df['hemisphere'] == "R", 'hemisphere'] = "r"
    out4 = utils.check_atlas_info(atlas['image'], atlas_df)
    pd.testing.assert_frame_equal(out, out4)

    # validate = True returns None
    none = utils.check_atlas_info(atlas['image'], atlas['info'], validate=True)
    assert none is None

    # providing labels allows for missing ids in atlas_info (i.e., does not
    # raise ValueError)
    drop_last_df = atlas_df.copy().iloc[:-1]
    out5 = utils.check_atlas_info(atlas['image'], drop_last_df,
                                  labels=range(1, 83))
    assert len(out5) == 82

    # not a filename or dataframe = failure
    with pytest.raises(TypeError):
        utils.check_atlas_info(atlas['image'], [1, 2, 3])

    # missing data = failure
    empty_df = pd.DataFrame(columns=['id', 'hemisphere', 'structure'])
    with pytest.raises(ValueError):
        utils.check_atlas_info(atlas['image'], empty_df)

    # invalid hemisphere designations
    bad_hemi_df = atlas_df.copy()
    bad_hemi_df.loc[1, 'hemisphere'] = 'notahemisphere'
    with pytest.raises(ValueError):
        utils.check_atlas_info(atlas['image'], bad_hemi_df)

    # invalid structural designation
    bad_struct_df = atlas_df.copy()
    bad_struct_df.loc[1, 'structure'] = 'notastructure'
    with pytest.raises(ValueError):
        utils.check_atlas_info(atlas['image'], bad_struct_df)


@pytest.mark.parametrize('metric, check, confirm, kwargs', [
    ('mean', np.array([[1, 2], [3, 4], [5, 6]]),
     np.array([1.5, 3.5, 5.5]), {'axis': 1}),
    ('median', np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
     np.array([2, 5, 8]), {'axis': 1}),
])
def test_check_metric(metric, check, confirm, kwargs):
    metric = utils.check_metric(metric)
    assert np.allclose(metric(check, **kwargs), confirm)

    # not an appropriate str
    with pytest.raises(ValueError):
        utils.check_metric('blargh')

    # doesn't accept axis kwarg
    with pytest.raises(TypeError):
        utils.check_metric(lambda x: np.mean(x))


@pytest.mark.xfail
def test_efficient_corr():
    assert False


@pytest.mark.xfail
def test_get_unique_labels():
    assert False


@pytest.mark.xfail
def test_get_centroids():
    assert False


@pytest.mark.xfail
def test_closest_centroid():
    assert False


@pytest.mark.parametrize('ijk, xyz', [
    ([0, 0, 0],
     np.array([[-90, -150, -80]])),
    ([[10, 10, 10], [100, 50, 100]],
     np.array([[-80, -140, -70], [10, -100, 20]])),
    ([[54, 32, 20], [82, 205, 38], [32, 51, 82]],
     np.array([[-36, -118, -60], [-8, 55, -42], [-58, -99, 2]])),
])
def test_coords_transform(ijk, xyz):
    affine = np.array([[1, 0, 0, -90],
                       [0, 1, 0, -150],
                       [0, 0, 1, -80],
                       [0, 0, 0, 1]])

    assert np.all(utils.xyz_to_ijk(xyz, affine) == ijk)
    assert np.all(utils.ijk_to_xyz(ijk, affine) == xyz)

    with pytest.raises(ValueError):
        utils.xyz_to_ijk([[10, 10], [20, 30]], affine)
    with pytest.raises(ValueError):
        utils.ijk_to_xyz([[10, 10], [20, 30]], affine)


@pytest.mark.xfail
def test_expand_roi():
    assert False



# -*- coding: utf-8 -*-
"""
Tests for abagen.probes module
"""

import numpy as np
import pandas as pd
import pytest

import abagen
from abagen import probes_
from abagen.utils import first_entry, flatten_dict


def test_reannotate_probes(testfiles):
    # set up a few useful variables
    probe_file = first_entry(testfiles, 'probes')

    # should work with either a filename _or_ a dataframe
    reannot = probes_.reannotate_probes(probe_file)
    probe_df = abagen.io.read_probes(probe_file)
    pd.testing.assert_frame_equal(reannot, probes_.reannotate_probes(probe_df))

    # expected output
    cols = ['probe_name', 'gene_symbol', 'entrez_id']
    assert np.all(reannot.columns == cols)
    assert reannot.index.name == 'probe_id'
    assert reannot.shape == (45821, 3)


@pytest.mark.parametrize('threshold, expected_length', [
    # threshold of zero should just return the full probe dataframe
    (0.0, 58692),
    # default threshold
    (0.5, 38176),
    # threshold of one should NOT return an empty dataframe (that's useless)
    # this will return all the probes that are greater than background noise
    # across ALL samples from ALL provided donors
    (1.0, 11878),
    # threshold is clipped to [0, 1] so these should be identical to [0, 1]
    (-1.0, 58692),
    (2.0, 11878),
])
def test_filter_probes(testfiles, threshold, expected_length):
    # set up a few useful variables
    pacall = flatten_dict(testfiles, 'pacall')
    probe_file = first_entry(testfiles, 'probes')
    samples = flatten_dict(testfiles, 'annotation')
    probe_df = abagen.io.read_probes(probe_file)

    # should work with either a filename _or_ a dataframe
    filtered = probes_.filter_probes(pacall, samples, probe_file,
                                     threshold=threshold)
    pd.testing.assert_frame_equal(
        filtered,
        probes_.filter_probes(pacall, samples, probe_df, threshold=threshold)
    )

    # provided threshold returns expected output
    cols = [
        'probe_name', 'gene_id', 'gene_symbol', 'gene_name', 'entrez_id',
        'chromosome'
    ]
    assert np.all(filtered.columns == cols)
    assert filtered.index.name == 'probe_id'
    assert len(filtered) == expected_length


@pytest.mark.parametrize('func, expected', [
    (probes_._max_idx, [3000, 5000, 6000]),
    (lambda x: x.index[0], [1000, 4000, 6000])
])
def test_groupby_and_apply(func, expected):
    # lots of set up for this little function...
    index = pd.Series([1000, 2000, 3000, 4000, 5000, 6000], name='probe_id')
    gene_symbol = ['a', 'a', 'a', 'b', 'b', 'c']
    info = pd.DataFrame(dict(gene_symbol=gene_symbol,
                             a=range(6),
                             b=range(12, 6, -1)),
                        index=index)
    prb = pd.DataFrame(dict(gene_symbol=gene_symbol), index=index)
    exp = [info[['a', 'b']].copy()]

    mi = list(probes_._groupby_and_apply(exp, prb, info, func).values())[0]
    pd.testing.assert_frame_equal(mi, info.loc[expected, ['a', 'b']].T)


def test_max_idx():
    df = pd.DataFrame(dict(a=[1, 2, 3],
                           b=[6, 5, 4]),
                      index=['one', 'two', 'three'])

    # should return max index of column specified
    assert probes_._max_idx(df, column='a') == 'three'
    assert probes_._max_idx(df, column='b') == 'one'

    # not specifying column should use first numerical column (i.e., 'a')
    assert probes_._max_idx(df) == 'three'

    # what if we only have a single column dataframe?
    dfa = df[['a']]
    assert probes_._max_idx(dfa, column='a') == 'three'
    assert probes_._max_idx(dfa) == 'three'

    # string column anywhere in dataframe is bad news bears
    with pytest.raises(TypeError):
        df['b'] = df['b'].astype(str)
        probes_._max_idx(df)


@pytest.mark.parametrize('angles, index', [
    (np.arange(0, 360, 15), 0),
    (np.arange(180, -180, -15), 12),
])
def test_max_loading(angles, index):
    def _get_ellipse_coords(ang, a=2, b=0.5, origin=(0, 0), rotate=0):
        # convert to radians
        rad = np.pi / 180
        ang, rotate = ang * rad, rotate * rad
        cang, sang = np.cos(ang), np.sin(ang)
        crot, srot = np.cos(rotate), np.sin(rotate)

        # calculate new points
        x = (a * cang * crot) - (b * sang * srot) + origin[0]
        y = (a * cang * srot) + (b * sang * crot) + origin[1]

        return x, y

    # grab a few points from a rotated ellipse and make a dataframe of it
    # the max loading should be wherever angle = 0
    df = pd.DataFrame([_get_ellipse_coords(f, rotate=45) for f in angles])
    assert probes_._max_loading(df) == index


@pytest.mark.parametrize('from_row, method, expected', [
    # when there are >2 rows the provided method is ignored
    (0, 'variance', 'one'),
    (0, 'intensity', 'one'),
    # when there are exactly two rows we use the specified method
    (1, 'variance', 'two'),
    (1, 'intensity', 'three'),
    # and one row just returns the index of that row
    (2, 'variance', 'three'),
    (2, 'variance', 'three'),
])
def test_correlate(from_row, method, expected):
    # a = max correlation, b = max variance, c = max intensity
    df = pd.DataFrame([[1.50, 2.00, 2.75],
                       [1.00, 2.00, 3.00],
                       [3.00, 3.00, 4.00]],
                      index=['one', 'two', 'three'], columns=['a', 'b', 'c'])

    assert probes_._correlate(df.iloc[from_row:], method=method) == expected


def test_diff_stability():
    # so much set up...
    index = pd.Series([1000, 2000, 3000, 4000, 5000, 6000], name='probe_id')
    prb = pd.DataFrame(dict(gene_symbol=['a', 'a', 'a', 'b', 'b', 'c']),
                       index=index)
    expression = [
        pd.DataFrame([[0, 2, 4, 9],
                      [1, 3, 3, 2],
                      [2, 4, 4, 1],
                      [3, 5, 1, 0],
                      [4, 6, 6, 7],
                      [5, 7, 3, 4]],
                     index=index, columns=['one', 'two', 'three', 'four']),
        pd.DataFrame([[1, 2, 3],
                      [3, 2, 1],
                      [1, 3, 2],
                      [1, 3, 2],
                      [1, 2, 3],
                      [1, 2, 3]],
                     index=index, columns=['one', 'two', 'three'])
    ]

    annotation = [
        pd.DataFrame(dict(structure_id=[0, 0, 1, 2]),
                     index=expression[0].columns),
        pd.DataFrame(dict(structure_id=[0, 1, 2]),
                     index=expression[1].columns)
    ]

    expected = [
        pd.DataFrame([[0, 2, 4, 9],
                      [4, 6, 6, 7],
                      [5, 7, 3, 4]],
                     index=pd.Series([1000, 5000, 6000], name='probe_id'),
                     columns=expression[0].columns).T,
        pd.DataFrame([[1, 2, 3],
                      [1, 2, 3],
                      [1, 2, 3]],
                     index=pd.Series([1000, 5000, 6000], name='probe_id'),
                     columns=expression[1].columns).T,
    ]

    out = list(probes_._diff_stability(expression, prb, annotation).values())
    assert len(out) == len(expected)
    for df, exp in zip(out, expected):
        pd.testing.assert_frame_equal(df, exp)


def test_average():
    index = pd.Series([1000, 2000, 3000, 4000, 5000, 6000], name='probe_id')
    prb = pd.DataFrame(dict(gene_symbol=['a', 'a', 'a', 'b', 'b', 'c']),
                       index=index)
    exp = [pd.DataFrame([[0, 12], [1, 11], [2, 10], [3, 9], [4, 8], [5, 7]],
                        index=index, columns=['a', 'b'])]

    # output should be a list of dataframes (only one dataframe in this case)
    out = probes_._average(exp, prb)
    assert len(out) == 1
    out = out[0]

    # confirm output dataframe is as expected
    expected = pd.DataFrame(dict(a=[1.0, 3.5, 5.0],
                                 b=[11.0, 8.5, 7.0]),
                            index=pd.Series(['a', 'b', 'c'],
                                            name='gene_symbol')).T
    pd.testing.assert_frame_equal(out, expected)


@pytest.mark.parametrize('method', [
    'average',
    'mean',
    'max_intensity',
    'max_variance',
    'pc_loading',
    'corr_intensity',
    'corr_variance',
    'diff_stability',
])
def test_collapse_probes(testfiles, method):
    # we've aleady tested the underlying methods so here we just want to do
    # some smoke tests to make sure the function returns what we expected
    # regardless of the provided method
    out = probes_.collapse_probes(flatten_dict(testfiles, 'microarray'),
                                  flatten_dict(testfiles, 'annotation'),
                                  first_entry(testfiles, 'probes'),
                                  method=method)

    out = list(out.values())
    assert len(out) == 2  # number of donors
    assert np.all([len(exp) == n_samp for exp, n_samp in zip(out, [363, 470])])
    assert np.all([len(exp.columns) == 29131 for exp in out])


@pytest.mark.parametrize('donor_probes', [
    'aggregate',
    'independent',
    'common'
])
def test_collapse_probes_donors(testfiles, donor_probes):
    out = probes_.collapse_probes(flatten_dict(testfiles, 'microarray'),
                                  flatten_dict(testfiles, 'annotation'),
                                  first_entry(testfiles, 'probes'),
                                  method='max_intensity',
                                  donor_probes=donor_probes)

    out = list(out.values())
    assert len(out) == 2  # number of donors
    assert np.all([len(exp) == n_samp for exp, n_samp in zip(out, [363, 470])])
    assert np.all([len(exp.columns) == 29131 for exp in out])



# -*- coding: utf-8 -*-

import os
import itertools

import nibabel as nib
import numpy as np
import pandas as pd
from scipy.ndimage import center_of_mass
from scipy.spatial.distance import cdist
from scipy.stats import zscore

AGG_FUNCS = dict(
    mean=np.mean,
    median=np.median
)


def check_dict(dictionary):
    """
    Checks that `dictionary` is a dict and makes it one, if not

    Parameters
    ----------
    dictionary : dict
        Presumptive dictionary

    Returns
    -------
    dictionary : dict
        Actual dictionary. If it wasn't one before, keys are chronological int,
        starting at 0
    """

    if isinstance(dictionary, str):
        dictionary = [dictionary]
    if not isinstance(dictionary, dict):
        return dict(zip(range(len(dictionary)), dictionary))
    return dictionary


def flatten_dict(dictionary, subkey):
    """
    Extracts `subkey` from entries of `dictionary` and creates new dictionary

    Parameters
    ----------
    dictionary : dict
        Input dictionary, where values are sub-dictionaries
    subkey : str
        Key to extract from `dictionary` entries

    Returns
    -------
    flattened : dict
        Flattened input `dictionary`

    Examples
    --------
    >>> test = {'one': {'value': 1}, 'two': {'value': 2}}
    >>> flatten_dict(test, 'value')
    {'one': 1, 'two': 2}
    """

    return {k: v.get(subkey, None) for k, v in check_dict(dictionary).items()}


def first_entry(dictionary, subkey=None):
    """
    Extracts `subkey` from the first entry in `dictionary`

    Parameters
    ----------
     dictionary : dict
        Input dictionary, where values are sub-dictionaries
    subkey : str
        Key to extract from `dictionary` entries

    Returns
    -------
    entry
        Extracted entry
    """

    dictionary = check_dict(dictionary)
    entry = dictionary[list(dictionary)[0]]
    if subkey is not None:
        entry = entry.get(subkey, None)

    return entry


def leftify_atlas(atlas):
    """
    Zeroes out all ROIs in the right hemisphere of `atlas`

    Assumes that positive X values indicate the right hemisphere (e.g., RAS+
    orientation) and that the X-origin is in the middle of the brain

    Parameters
    ----------
    atlas : str or niimg-like
        Filepath to or in-memory loaded image

    Returns
    -------
    atlas : niimg-like
        Loaded image with right hemisphere zeroed out
    """

    atlas = check_img(atlas)

    # get ijk corresponding to zero-point
    i, j, k = nib.affines.apply_affine(np.linalg.inv(atlas.affine), [0, 0, 0])

    # zero out all positive voxels; img is RAS+ so positive = right hemisphere
    data = np.array(atlas.dataobj, copy=True)
    data[int(i):] = 0

    return atlas.__class__(data, atlas.affine, header=atlas.header)


def check_img(img):
    """
    Very basic checker that loads `img`` and ensures it's 3D/int

    Parameters
    --------
    img : str or niimg-like
        Filepath to or in-memory loaded image

    Returns
    -------
    img : niimg-like
        Loaded 3D/int image
    """

    if isinstance(img, (str, os.PathLike)) and os.path.exists(img):
        img = nib.load(img)
    elif not isinstance(img, nib.spatialimages.SpatialImage):
        raise TypeError('Provided image must be an existing filepath or a '
                        'pre-loaded niimg-like object')

    # ensure 3D or squeezable to 3D
    if len(img.shape) == 4 and img.shape[3] == 1:
        data = np.asarray(img.dataobj)
        affine = img.affine
        img = img.__class__(data[:, :, :, 0], affine, header=img.header)
    elif len(img.shape) != 3:
        raise ValueError('Provided image must be 3D')

    # check if atlas data is int or castable to int
    # if image is arrayproxy convert it to an array for speed-up
    data = np.asarray(img.dataobj)
    cast = nib.is_proxy(img.dataobj)
    if img.header.get_data_dtype().kind not in ['i', 'u']:
        idata = data.astype('int32')
        cast = np.allclose(idata, data)
        data = idata
        if not cast:
            raise ValueError('Provided image should have integer values or '
                             'be safely castable to int without data loss')
    if cast:
        img = img.__class__(data, img.affine, header=img.header)
        img.header.set_data_dtype(np.int32)

    return img


def check_atlas_info(atlas, atlas_info, labels=None, validate=False):
    """
    Checks whether provided `info` on `atlas` is sufficient for processing

    Parameters
    ----------
    atlas : niimg-like object
        Parcellation image, where voxels belonging to a given parcel should be
        identified with a unique integer ID
    atlas_info : str or pandas.DataFrame
        Filepath or dataframe containing information about `atlas`. Must have
        at least columns 'id', 'hemisphere', and 'structure' containing
        information mapping atlas IDs to hemisphere (i.e., "L" or "R") and
        broad structural class (i.e.., "cortex", "subcortex", "cerebellum",
        "brainstem", "white matter", or "other").
    labels : array_like, optional
        List of values containing labels to compare between `atlas` and
        `atlas_info`, if they don't all match. If not specified this function
        will attempt to confirm that all IDs present in `atlas` have entries in
        `atlas_info` and vice versa. Default: None
    validate : bool, optional
        Whether to only validate (True) the provided `atlas` and `atlas_info`
        instead of returning (False) the validated dataframe. Default: False

    Returns
    -------
    atlas_info : pandas.DataFrame
        Loaded dataframe with information on atlas
    """

    from .samples_ import ONTOLOGY  # avoid circular imports

    atlas = check_img(atlas)
    ids = get_unique_labels(atlas) if labels is None else labels
    valid_structures = list(ONTOLOGY.value_set('structure'))
    hemi_swap = {
        'lh': 'L', 'LH': 'L', 'l': 'L',
        'rh': 'R', 'RH': 'R', 'r': 'R'
    }
    expected_cols = ['hemisphere', 'structure']

    # load info, if not already
    if not isinstance(atlas_info, pd.DataFrame):
        try:
            atlas_info = pd.read_csv(atlas_info)
        except (ValueError, TypeError):
            pass

    try:
        atlas_info = atlas_info.copy()
        if 'id' in atlas_info.columns:
            atlas_info = atlas_info.set_index('id')
    except AttributeError:
        raise TypeError('Provided atlas_info must be a filepath or pandas.'
                        'DataFrame. Please confirm inputs and try again.')

    try:
        assert all(c in atlas_info.columns for c in expected_cols)
        assert 'id' == atlas_info.index.name
        assert len(np.setdiff1d(ids, atlas_info.index)) == 0
    except AssertionError:
        raise ValueError('Provided atlas_info does not have adequate '
                         'information on supplied atlas. Please confirm '
                         'that atlas_info has columns [\'id\', '
                         '\'hemisphere\', \'structure\'], and that the region '
                         'IDs listed in atlas_info account for all those '
                         'found in atlas.')

    try:
        atlas_info['hemisphere'] = atlas_info['hemisphere'].replace(hemi_swap)
        hemi_diff = np.setdiff1d(atlas_info['hemisphere'], ['L', 'R'])
        assert len(hemi_diff) == 0
    except AssertionError:
        raise ValueError('Provided atlas_info has invalid values in the'
                         '\'hemisphere\' column. Only the following values '
                         'are allowed: {}. Invalid value(s): {}'
                         .format(['L', 'R'], hemi_diff))

    try:
        struct_diff = np.setdiff1d(atlas_info['structure'], valid_structures)
        assert len(struct_diff) == 0
    except AssertionError:
        raise ValueError('Provided atlas_info has invalid values in the'
                         '\'structure\' column. Only the following values are '
                         'allowed: {}. Invalid value(s): {}'
                         .format(valid_structures, struct_diff))

    if not validate:
        return atlas_info


def check_metric(metric):
    """
    Confirms `metric` is a valid aggregation metric

    Parameters
    ----------
    metric : {'mean', 'median'} or callable, optional
        Mechanism by which to reduce arrays. If a callable, should be able to
        accept an `N`-dimensional input and the `axis` keyword argument and
        return an `N-1`-dimensional output. Default: 'mean'

    Returns
    -------
    metric : callable
    """

    if not callable(metric):
        try:
            metric = AGG_FUNCS[metric]
        except KeyError:
            raise ValueError('Provided metric {0} is not valid. If supplied'
                             'as string, metric must be in {1}.'
                             .format(metric, list(AGG_FUNCS)))
    else:
        try:
            test = np.random.rand(10, 10, 10)
            assert metric(test, axis=0).shape == (10, 10)
            assert isinstance(metric(test), float)
        except (AssertionError, TypeError):
            raise TypeError('Provided metric {0} does not perform as '
                            'expected. Please ensure it accepts the `axis` '
                            'keyword argument and can reduce an `N`-'
                            'dimensional input to an `N-1`-dimensional output.'
                            .format(metric))

    return metric


def efficient_corr(x, y):
    """
    Computes correlation of matching columns in `x` and `y`

    Parameters
    ----------
    x, y : (N, M) array_like
        Input data arrays

    Returns
    -------
    corr : (M,) numpy.ndarray
        Correlations of columns in `x` and `y`
    """

    corr = np.sum(zscore(x, ddof=1) * zscore(y, ddof=1), axis=0) / (len(x) - 1)

    return corr


def get_unique_labels(label_image):
    """
    Returns all possible ROI labels from ``label_image``

    Parameters
    ----------
    label_image : niimg-like object
        ROI image, where each ROI is identified with a unique integer ID

    Returns
    -------
    labels : np.ndarray
        Integer labels of all ROIS found within ``label_image``
    """

    label_image = check_img(label_image)
    return np.trim_zeros(np.unique(label_image.dataobj)).astype(int)


def get_centroids(image, labels=None, image_space=False):
    """
    Finds centroids of ``labels`` in ``label_image``

    Parameters
    ----------
    label_image : niimg-like object
        3D image containing integer label at each point
    labels : array_like, optional
        List of values containing labels of which to find centroids. Default:
        all possible labels
    image_space : bool, optional
        Whether to return xyz (image space) coordinates for centroids based
        on transformation in ``label_image.affine``. Default: False

    Returns
    -------
    centroids : (N, 3) np.ndarray
        Coordinates of centroids for ROIs in input data
    """

    image = check_img(image)
    data = np.asarray(image.dataobj)

    # if no labels of interest provided, get all possible labels
    if labels is None:
        labels = np.trim_zeros(np.unique(data))

    # get centroids for all possible labels
    centroids = np.row_stack(center_of_mass(data, labels=data, index=labels))

    # return xyz if desired; otherwise, ijk
    if image_space:
        centroids = ijk_to_xyz(centroids, image.affine)

    return centroids


def closest_centroid(coord, centroids, return_dist=False):
    """
    Returns index of `centroids` closest to `coord`

    Computes Euclidean distances between `coord` and each of `centroids`,
    returning index of closest centroid

    Parameters
    ----------
    coord : (1, 3) array_like
        Coordinates of sample
    centroids : (N, 3) array_like
        Centroids of parcels (in same space as `coord`)
    return_dist : bool, optional
        Whether to also return distance of closest centroid

    Returns
    -------
    closest : int
        Index of closest centroid in `centroids` to `coord`
    distance : float
        Distance of closest centroid in `centroids` to `coord`. Only returned
        if `return_dist=True`
    """

    distances = np.squeeze(cdist(np.atleast_2d(coord), centroids))
    closest = distances.argmin(axis=0)

    if return_dist:
        return closest, distances[closest]

    return closest


def _check_coord_inputs(coords):
    """
    Confirms `coords` are appropriate shape for coordinate transform

    Parameters
    ----------
    coords : array-like

    Returns
    -------
    coords : (3, N) np.ndarray
    """

    coords = np.atleast_2d(coords).T
    if 3 not in coords.shape:
        raise ValueError('Input coordinates must be of shape (3 x N). '
                         'Provided coordinate shape: {}'.format(coords.shape))
    if coords.shape[0] != 3:
        coords = coords.T
    # add constant term to coords to make 4 x N
    coords = np.row_stack([coords, np.ones_like(coords[0])])
    return coords


def ijk_to_xyz(coords, affine):
    """
    Converts `coords` in cartesian space to `affine` space

    Parameters
    ----------
    coords : (N, 3) array_like
        Cartesian (ijk) coordinate values
    affine : (4, 4) array_like
        Affine matrix containing displacement + boundary

    Returns
    -------
    xyz : (N, 3) np.ndarray
        Provided ``coords`` in ``affine`` space
    """

    coords = _check_coord_inputs(coords)
    aff_coords = np.dot(affine, coords)[:3].T
    return aff_coords


def xyz_to_ijk(coords, affine):
    """
    Converts `coords` in `affine` space to cartesian space

    Parameters
    ----------
    coords : (N, 3) array_like
        Image coordinate (xyz) values
    affine : (4, 4) array_like
        Affine matrix containing displacement + boundary

    Returns
    ------
    ijk : (N, 3) numpy.ndarray
        Provided `coords` in cartesian space
    """

    coords = _check_coord_inputs(coords)
    ijk_coords = np.linalg.solve(affine, coords)[:3].T.astype(int)
    return ijk_coords


def expand_roi(coords, dilation=1, return_array=True):
    """
    Expands coordinates ``coords`` to include neighboring coordinates

    Computes all possible coordinates of distance ``dilation`` from ``coords``.
    Returns a generator (``itertools.product``) by default, but can return an
    array if desired (if ``return_array=True``).

    Parameters
    ----------
    coords : (1, 3) array_like
        List of ijk values for coordinate in 3D space
    dilation : int, optional
        How many neighboring voxels to expand around `coords`. Default: 1
    return_array : bool, optional
        Whether to return array instead of generator. Default: True

    Returns
    -------
    coords : (27, 3) generator
        Coordinates of expanded ROI
    """

    def expand(x, d=1):
        return np.arange((x - d), (x + d + 1), dtype=int)

    # return all combinations of coordinates
    coords = np.squeeze(coords)
    gen = itertools.product(*[expand(n, d=dilation) for n in coords])

    # coerce to array if desired
    if return_array:
        return np.asarray(list(gen))

    return gen
